'''
Parameter class and function for Riverdale
This is the user-side parameter sheet, 
not to be confused with the singleton parameter classes managing the compile time constants located at the top of some files

To some exceptions, this parameter class manages all the inputs to the model (e.g. initial topo, uplift, precipitations, manning, ...)

Authors:
	- B.G.
'''

import numpy as np
from enum import Enum
import scabbard.riverdale.rd_grid as rdgd
import scabbard.riverdale.rd_morphodynamics as rdmo
import scabbard.riverdale.rd_hydrodynamics as rdhy
import scabbard.utils as scaut 
import scabbard as scb
import dagger as dag
import warnings
import taichi as ti
import sys, os


# Disable
def blockPrint():
	sys.stdout = open(os.devnull, 'w')

# Restore
def enablePrint():
	sys.stdout = sys.__stdout__

def reinit_ti(rd, params):
	blockPrint()
	QwA = rd.QwA.to_numpy()
	hw = rd.hw.to_numpy()
	# params = rd.param
	ti.init(ti.gpu)
	# print(params)
	params.rd = None
	rd = scb.rvd.create_from_params(params)
	rd.QwA.from_numpy(QwA)
	rd.hw.from_numpy(hw)
	enablePrint()
	return rd


class RDParams:
	'''
	Parameter class for RiverDale, contains the sheet of parameters required Riverdale's model
	This is the user-side parameter sheet, managing the precipitation, roughness, erodability, critical shear stress ...

	It manages inputs for the model, some of them are to be fixed before initiating the model and others can be edited during the model lifetime.

	Authors:
		- B.G (last modification on the 30/04/1992)
	'''


	############################################################
	# CONSTRUCTOR ##############################################
	############################################################


	def __init__(self):
		'''
			The constructor does not actually take parameters: it initiates all the parameters to their default values
			Note, internal variables starts with _. Do not use them directly if you do not know what you are doing.
			Most of the API is exposed via cleaner properties (setters and getters) without the _ (see all the @property) or direct functions

			Returns:
				- A default RDParams parameter sheet
			Authors:
				- B.G. (last modifications: 29/05/2024)
		'''

		# _RD is a reference to the model object (RiverDale)
		# Once initialised, RD also gets a ref to this param sheet
		# Constant parameters becomes lock and the variable inputs will be automatically transmitted on modification
		self._RD = None

		##############################
		# GRID Parameters
		##############################

		# Grid Geometry, probably no need to expose them directly:
		## spatial step in the x direction
		self._dx = 1.
		## spatial step in the y direction (Warning, some of the codes only considers dx=dy)
		self._dy = 1.
		## Number of columns
		self._nx = 50
		## Number of rows
		self._ny = 50
		## Number of nodes in total
		self._nxy = self._nx * self._ny


		self.dtype_float = ti.f32

		##############################
		# Boundary conditions
		##############################
		
		# Type of boundary conditions for the flow connectivity and no data management
		## This enum is defined in rd_grid.py and determines how the codes reads/deal bcs
		self._boundaries = rdgd.BoundaryConditions.normal
		## Arrays of boundary conditions - used if the boundary conditions are set to customs
		## See rd_grid.py for the meaning (or the future doc at the date of the 29/05/2024)
		self._BCs = None

		# Slope boundary conditions: what slope to apply on outting cells
		self._boundary_slope_mode = rdgd.BoundaryConditionsSlope.fixed_slope
		self._boundary_slope_value = 1e-2

		# Can the nodes who input Qs/Qw (boundaries GIVE or FORCE_GIVE) update their topo in morphomode
		self._update_morpho_at_input_points = False


		##############################
		# Initial conditions
		##############################

		# Initial topography, fed to the model at initiation and also kept here for comparisons or things like that
		self.initial_Z = np.random.rand(self._ny, self._nx).astype(dtype = np.float32)
		
		# Initial flow depth - kept to 0 if None
		self.initial_hw = None

		
		##############################
		# HYDRO Parameters
		##############################

		self._hydro_compute_mode = rdhy.FlowMode.static_incremental
		self._use_fdir_D8  = True

		
		# Manning's Roughness/friction coefficient
		self._manning = 0.033
		
		# time step for the hydrodynamics model
		self._dt_hydro = 1e-3
		
		self._LM_npath = 1
		self._clamp_div_hw_val = 1e-3

		# Precipitation rates
		## The value (2 or 2D)
		## Once the model is initialised, it cannot be switched from 1D to 2D
		self._precipitations = 50*1e-3/3600
		## Does the model even need to run the precipitation input?
		self._has_precipitations = True
		## Is the precipitation field field in 2D
		self._2D_precipitations = False
		## are precipitations manually cut off
		self._force_no_prec = False

		# Discrete input points
		# Default is no direct inputs
		# The models checks if this is required with the need_input_Qw property
		## Indices of the row inputs
		self._input_rows_Qw = None
		## Indices of the columns inputs
		self._input_cols_Qw = None
		## Input values in m^3/s
		self._input_Qw = None

		self._precompute_Qw = False


		self._stationary = True

		#TODO once I find a stable method
		# # Stability criterion
		# self._



		##############################
		# MORPHO Parameters
		##############################


		# Is the Morphodynamics module enabled?
		## If False, it does not run a single morpho functions and you can ignore the params
		self._morpho = False

		# Time step for the morphodynamics modelling		
		self._dt_morpho = 1e-3

		# Enum class for the type of fluvial process activated (Ignore, so far just one)
		self._morphomode = rdmo.MorphoMode.fbal

		# Gravitational constant g
		self._GRAVITY = 9.81

		# Water density
		self._rho_water = 1000

		# Sediment density
		self._rho_sediment = 2600

		# Gravitational erosion coeff
		self._k_z = 1.

		# shear-stress entrainement local coeff erosion coeff
		self._k_h = 1.

		# Fluvial erosion coeff (the dimentional E in MPM)
		self._k_erosion = 1e-5

		# Fluvial exponent
		self._alpha_erosion = 1.5

		# Grainsize
		self._D = 4e-3 

		# Critical shear stress
		self._tau_c = 4

		# critical shear strass
		self._transport_length = 4

		
	############################################################
	# GRID Stuffs ##############################################
	############################################################


	def set_initial_conditions(self, nx, ny, dx, dy, Z, hw = None, BCs = None):
		'''
		Set up the initial conditions of the model, from the grid to the Z, BCs or eventually an initial flow depth
		Needs to be done before the model setup

		Arguments:
			- nx: number of columns in the grid
			- ny: number of rows in the grid
			- dx: spatial step in the x direction
			- dy: spatial step in the y direction
			- Z: a numpy array of 2D initial elevation 
			- hw: an optional array of initial flow depth
			- BCs: an optional array of Boundary conditions code in uint8
		returns:
			- Nothing but sets up the inital conditions of the model
		Authors:
			- B.G (last modifications: 30/04/2024)
		TODO:
			- add checks on the dimensions and boundary conditions, also warning is BCs is given but boundaries are set to another mode
		'''

		if(self._RD is not None):
			raise Exception("Cannot modify initial condition and grid parameters after model initiation")

		if (dx != dy):
			raise Exception("Differential spatial step in x and y in WIP, I am even considering dropping it, let me know if needed")

		self._dx = dx
		self._dy = dy
		self._nx = nx
		self._ny = ny
		self._nxy = self._nx * self._ny
		self._boundaries = rdgd.BoundaryConditions.normal
		self.BCs = BCs
		self.initial_Z = Z
		self.initial_hw = hw if(hw is not None) else np.zeros_like(self.initial_Z)


	@property
	def boundaries(self):
		'''
			Boundary enum code of the model, see BoundaryConditions enum class in rd_grid.py
			Authors:
				- B.G (last modification: 29/05/2024)
		'''
		return self._boundaries

	@boundaries.setter
	def boundaries(self, val):
		'''
			Setter for changing the type of boundary conditions in the model.
			It cannot be changed once the model is instantiated.
			Note that if the boundaries are set to `customs`, you can change the boundary codes by modifying BCs.

			Authors:
				- B.G. (last modification: 29/05/2024)	
		'''
		if(val == rdgd.BoundaryConditions.normal):
			self._boundaries = val
		elif(val == rdgd.BoundaryConditions.customs):
			raise ValueError("To set the boundary conditions to custom values in the parameter sheet, you need to directly provide a 2D numpy array of np.uint8 codes to the member param.BCs (where param is the parameter sheet object)")
		else:
			raise NotImplementedError('Boundary condition not implemented yet')

	@property
	def BCs(self):
		'''
		 Actual boundary codes used in the case of customs boundary conditions 
		 See the convention in rd_grid.py
		 Authors:
		 	B.G. (last modification: 29/05/2024)
		'''
		return self._BCs

	@BCs.setter
	def BCs(self, val):
		'''
			Setter for the boundary condition codes.
			Needs to be an uint8 nupy array following the convention in rd_grid.py.
			Note that setting this values to a 2D numpy array sets the boundary condition type to customs
			Authors:
				- B.G (29/05/2024)
		'''
		# if the value is None there is not much to do
		if val is None:
			return

		 # Sanitising input
		if not isinstance(val, np.ndarray) or not len(val.shape) == 2 or not val.shape[0] == self._ny or not val.shape[1] == self._nx or not val.dtype == np.uint8:
			raise ValueError(f"If you want to directly set customs boundary conditions, please provide a 2D array of shape {self._ny},{self._nx} of type np.uint8")

		# Registering it if the model is not instanciated
		if(self._RD is None):
			self._BCs = val
			self._boundaries = rdgd.BoundaryConditions.customs

		# If the model is instanciated I need extra checks and work
		else:
			# Checking if the mdel has been initialised to the right mode
			if(self.boundaries != rdgd.BoundaryConditions.customs):
				raise ValueError('Model is already initialised in another boundary mode than customs. BCs cannot be modified.')
			else:
				# Directly transferring data to the GPU BCs code
				self._RD.BCs.from_numpy(val)


	def set_boundary_slope(self, val, mode = 'slope'):
		'''
		Set the mode of computation for the slope at the boundary of the model, when can_out is True.
		As of today, there are two options: Fixing a slope at the boundary or fixing a receiver elevation.
		Arguments:
			- val: the value at the boundary (either the slope or the elevation)
			- mode: 'slope' or 'elevation'
		returns:
			- Nothing
		Authors:
			- B.G. (last modification: 30/05/2024)
		'''

		# Function can only be executed if RD not initialised
		if(self._RD is not None):
			raise ValueError("Cannot change the boundary value for slope calculation if the model is already instantiated")

		# Case of fixed slope at boundary can_out
		if(mode == 'slope'):
			# Setting the internal enum
			self._boundary_slope_mode = rdgd.BoundaryConditionsSlope.fixed_slope
			# Slope cannot be negative
			if(val < 0):
				warnings.warn('Recasting negative slope to 0 for the boundary condition slope calculation')
				val = 0.
		# case of fixed elevation of virtual receiver
		elif(mode == 'elevation'):
			# Setting the internal enum
			self._boundary_slope_mode = rdgd.BoundaryConditionsSlope.fixed_elevation
		# Other potential methods not implemented yet
		else:
			raise ValueError(f'mode needs to be "slope" or "elevation", it cannot be {mode}')

		# Actually setting the value
		self._boundary_slope_value = val
		


	@property
	def hydro_compute_mode(self):
		return self._hydro_compute_mode

	@property
	def force_LM_to_original_flowdir(self):
		return self._force_LM_to_original_flowdir

	@hydro_compute_mode.setter
	def hydro_compute_mode(self, value):
		'''

		'''

		# # At the moment it needs to be compile-time constant
		# if(self._RD is not None):
		# 	raise Exception("Compute mode for hydro cannot be changed while the model is already initialised (WIP to dynamically do that)")
		# else:
		self._hydro_compute_mode = value
		if(self._RD is not None):
			self._RD.PARAMHYDRO.flowmode = value
			rdhy.set_hydro_CC()
			if (self._RD.constraints is None and rdhy.FlowMode.static_drape == self._RD.param._hydro_compute_mode):
				self._RD.constraints = ti.field(self._RD.param.dtype_float, shape = (self._RD.GRID.ny,self._RD.GRID.nx,2))



	@property
	def manning(self):
		'''
			Returns the roughness coefficient for the friction equation used to calculate flow velocity
			Authors:
				- B.G. (last modification 30/04/2024)
		'''
		return self._manning

	@manning.setter
	def manning(self, value):
		'''
			Setter function for the roughness coefficient, essentially checking if the input is 1 or 2D
			Authors:
				- B.G. (last modification 30/04/2024)
		'''

		# here I'll lay out the code for 2D mannings
		if(isinstance(value,np.ndarray)):
			raise Exception("2D Roughness coefficient not supported yet")

		# At the moment it needs to be compile-time constant
		if(self._RD is not None):
			raise Exception("Roughness coefficient cannot yet be modified after setting up the model. Working on a robust way to do so but so far it is a compile-time constant for performance reasons and therefore cannot be changed after creating the model instance")
		else:
			self._manning = value

	@property
	def exponent_flow(self):
		return rdhy.PARAMHYDRO.exponent_flow

	@exponent_flow.setter
	def exponent_flow(self, value):
		# At the moment it needs to be compile-time constant
		if(self._RD is not None):
			raise Exception("Flow exponent for the friction equation cannot yet be modified after setting up the model. Working on a robust way to do so but so far it is a compile-time constant for performance reasons and therefore cannot be changed after creating the model instance")
		else:
			rdhy.PARAMHYDRO.exponent_flow = value

	@property
	def dt_hydro(self):
		'''
			Returns the time step used in the simulation for hydrodynamics
			Authors:
				- B.G. (last modification 30/04/2024)
		'''
		return self._dt_hydro

	@dt_hydro.setter
	def dt_hydro(self, value):
		'''
			Setter function for the time step used in the simulation for hydrodynamics
			nothing particular but is future proofed
			Authors:
				- B.G. (last modification 30/04/2024)
		'''

		self._dt_hydro = value

		if not (self._RD is None):
			raise ValueError("cannot change dt_hydro after initialisation (static variable). You need to call the dedicated function that reboot the model")

	@property
	def precipitations(self):
		'''
			Returns the field of precipitation rates (whether it si 1D or 2D)
			Authors:
				- B.G. (last modification 30/04/2024)
		'''
		return self._precipitations

	@precipitations.setter
	def precipitations(self, value):
		'''
			Setter function for the precipitation rates, essentially checking if the input is 1 or 2D
			Note: this value can be modified even if the model is already instantiated/running

			Arguments:
				- value: a scalar or 2D array of domain size of precipitation rates in m/s
			Returns:
				- Nothing, update the parm sheet and the model in place
			Authors:
				- B.G. (last modification 23/05/2024)
		'''

		#
		if(self._force_no_prec):
			warnings.warn('Precipitations have been manually disabled using the method `disable_precipitations`, new values will not be taken into account unless the method `enable_precipitations` is run later on.')

		# Checking if the input is 1 or 2D
		if(isinstance(value,np.ndarray)):
			input2D = True
		else:
			input2D = False

		# Checking if the precipitation module is even needed (e.g. 0 precipitations)
		if(not input2D and value == 0):
			self._has_precipitations = False
		else:
			self._has_precipitations = True

		# Cannot change the type of input if the runtime is already initiated
		if(self._2D_precipitations != input2D and self._RD is not None):
			raise Exception("You cannot change precipitations from 1D to 2D (or the other way) once the model is instantiated (working on that). (scb.rvd.change_hydro_dt)")

		# if I reach here I can set the parameters
		self._precipitations = value

		# Apply the modification to the model in the case it is already instantiated
		if(self._RD is not None):
			if(isinstance(value,np.ndarray)):
				value = value.astype(np.float32)
				self._RD.P.from_numpy(value)
			else:
				value = np.float32(value)
				self._RD.P = value

	def set_input_Qw(self, rows, cols, values, disable_precipitations = False):
		'''
		Function to set the discrete input points for the hydrodynamics model.
		Note that it replaces any previous discrete inputs, but it does not replace the precipitations if there are any
		Arguments:
			- rows: indices of the rows where the inputs flow in
			- cols: indices of the columns where the inputs flow in
			- values: input Qw in m^3/s
			- disable_precipitations: switch off any precipitation inputs if true
		Returns:
			- NOthing, update the param sheet in place
		'''
		# couple of checks first on the data format
		if scaut.is_numpy(values) == False or scaut.is_numpy(rows, dtype = np.integer) == False or scaut.is_numpy(cols, dtype = np.integer) == False:
			print(scaut.is_numpy(values) ,'||', scaut.is_numpy(rows, dtype = np.integer) ,'||', scaut.is_numpy(cols, dtype = np.integer))
			raise ValueError('rows and cols must be 1D numpy arrays of integers and values a numpy array of float')

		# Checking the dimensions too
		tgshape = values.shape
		## If they are ot of the same dimension, it would crash the model
		if(values.shape != tgshape or rows.shape != tgshape or cols.shape != tgshape):
			raise ValueError('Inputs arrays must have the same shape')

		# Setting the data in the param sheet
		self._input_rows_Qw = rows
		self._input_cols_Qw = cols
		self._input_Qw = values

		# Updating the data in the model in case is already running
		if(self._RD is not None):	
			self._RD.input_rows_Qw.from_numpy(rows)
			self._RD.input_cols_Qw.from_numpy(cols)
			self._RD.input_Qw.from_numpy(values)

		# Done


	@property
	def need_precipitations(self):
		'''
		Property determining if I need to run the pieces of code adding the precipitation or not
		returns True or False
		Authors:
				- B.G. (last modification 23/05/2024)
		'''
		return self._has_precipitations and self._force_no_prec == False

	@need_precipitations.setter
	def need_precipitations(self, val):
		'''
		Explicitely forbidding the manual assignement for this property
		It is computed from the various inputs or more explicit function
		Authors:
			- B.G. (last modification 23/05/2024)
		'''
		print("Cannot manually set need_precipitations")
	
	@property
	def precipitations_are_2D(self):
		# Checking if the input is 1 or 2D
		if(isinstance(self.precipitations, np.ndarray)):
			input2D = True
		else:
			input2D = False
		return input2D and self.need_precipitations

	def disable_precipitations(self):
		'''
		Explicitely disable precipitations
		Authors:
				- B.G. (last modification 23/05/2024)
		'''
		self._force_no_prec = True

	def enable_precipitations(self):
		'''
		Explicitely re-enable precipitations (only required if it has been explicitely disabled before)
		Authors:
				- B.G. (last modification 23/05/2024)
		'''
		self._force_no_prec = False


	@property
	def need_input_Qw(self):
		'''
		Determines whether the model needs to run the discrete input function or not.
		Authors:
			- B.G (last modification: 28/05/2024)
		'''
		if(self._input_rows_Qw is not None):
			return True
		else:
			return False

	def preflood_lakes(self, fill_coeff = 0.75):

		tZ = self.initial_Z.copy()
		if(not (self.initial_hw is None)):
			tZ += self.initial_hw
		diff = np.copy(tZ)
		scb.flow.priority_flood(tZ, BCs = scb.flow.get_normal_BCs(tZ) if self.BCs is None else self.BCs, D4 = True, in_place = True, dx = 1., step_fill = 1e-3)
		diff = (tZ - diff) * fill_coeff

		self.initial_Z += diff

		return diff


	@property
	def update_morpho_at_input_points(self):
		return self._update_morpho_at_input_points

	@update_morpho_at_input_points.setter
	def update_morpho_at_input_points(self, val:bool):
		if(self._RD is None):
			self._update_morpho_at_input_points = val
		else:
			warnings.warn('Cannot update static update_morpho_at_input_points param after model initialisation')

	@property
	def use_fdir_D8(self):
		'''
		Authors:
				- B.G. (last modification 09/2024)
		'''
		return self._use_fdir_D8

		
	@use_fdir_D8.setter
	def use_fdir_D8(self, val:bool):
		if(self._RD is None):
			self._use_fdir_D8 = val
		else:
			warnings.warn('Cannot update static use_fdir_D8 param after model initialisation')

	@property
	def stationary(self):
		'''
		if True, model runs in stationary mode, i.e. it does not propagate flood wave but fous on the equillibirum solution
		If False, propagate transient flow through time (warning, this is much slower)
		'''
		return self._stationary

	@stationary.setter
	def stationary(self,val:bool):
		self._stationary = val




	@property
	def reshp(self):
		return self.ny,self.nx

	@property
	def precompute_Qw(self):
		return self._precompute_Qw

	@precompute_Qw.setter
	def precompute_Qw(self,val):
		self._precompute_Qw = val

	@property
	def dt_morpho(self):
		return self._dt_morpho

	@dt_morpho.setter
	def dt_morpho(self,val):
		self._dt_morpho = val

	@property
	def morpho(self):
		return self._morpho

	@morpho.setter
	def morpho(self,val):
		self._morpho = val

	@property
	def morphomode(self):
		return self._morphomode

	@morphomode.setter
	def morphomode(self,val):
		self._morphomode = val

	@property
	def GRAVITY(self):
		return self._GRAVITY

	@GRAVITY.setter
	def GRAVITY(self,val):
		self._GRAVITY = val

	@property
	def rho_water(self):
		return self._rho_water

	@rho_water.setter
	def rho_water(self,val):
		self._rho_water = val

	@property
	def rho_sediment(self):
		return self._rho_sediment

	@rho_sediment.setter
	def rho_sediment(self,val):
		self._rho_sediment = val

	@property
	def k_z(self):
		return self._k_z

	@k_z.setter
	def k_z(self,val):
		self._k_z = val

	@property
	def k_h(self):
		return self._k_h

	@k_h.setter
	def k_h(self,val):
		self._k_h = val

	@property
	def k_erosion(self):
		return self._k_erosion

	@k_erosion.setter
	def k_erosion(self,val):
		self._k_erosion = val

	@property
	def alpha_erosion(self):
		return self._alpha_erosion

	@alpha_erosion.setter
	def alpha_erosion(self,val):
		self._alpha_erosion = val

	@property
	def D(self):
		return self._D

	@D.setter
	def D(self,val):
		self._D = val

	@property
	def tau_c(self):
		return self._tau_c

	@tau_c.setter
	def tau_c(self,val):
		self._tau_c = val

	@property
	def transport_length(self):
		return self._transport_length

	@transport_length.setter
	def transport_length(self,val):
		self._transport_length = val




def param_from_grid(grid):
	'''
	TODo
	'''
	param = RDParams()
	if(isinstance(grid,scb.raster.RegularRasterGrid)):
		param.set_initial_conditions(grid.geo.nx, grid.geo.ny, grid.geo.dx, grid.geo.dx, grid.Z)
	else:
		param.set_initial_conditions(grid.nx, grid.ny, grid.dx, grid.dx, grid.Z2D)
	return param

def param_from_dem(dem, initial_fill = True):
	return param_from_grid(scb.io.load_raster(dem))



def change_dt_hydro(rd,param, new_dt):
	'''
	Recompile the model with a new dt_hydro. Required as this parameter is a compile-time constant

	Arguments:
		- rd: the riverdale model
		- param: the param object
		- new_dt: the new time step for the hydro model

	Returns:
		- a new rd object, edits param in place

	Authors: 
		- B.G. (last modification: 09/2024)
	'''
	param._dt_hydro = new_dt
	return reinit_ti(rd, param)


'''
Sets of function to compute hydrodynamics with RiverDale

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_helper_surfw as hsw
import scabbard.riverdale.rd_hydrodynamics as hydrofunc



class MorphoMode(Enum):
	'''
	Enumeration of the different boundary condition types possible
	'''	
	fbal = 0,
	MPM_like = 1,
	fbal_links = 2


@scaut.singleton
class MorphoParams:
	'''
		Internal singleton class holding all the compile time constant parameters for the hydro simulations 
		Not for users
	'''
	def __init__(self):

		# Time for the morphodynamics modelling		
		self.dt_morpho = 1e-3
		# What kind of erosion
		self.morphomode = MorphoMode.fbal
		# gravitational constant
		self.GRAVITY = 9.81
		# Water density
		self.rho_water = 1000
		# Sediment density
		self.rho_sediment = 2600
		# Gravitational erosion coeff
		self.k_z = 1.
		self.k_h = 1.
		# Fluvial erosion coeff
		self.k_erosion = 1e-5
		# Fluvial exponent
		self.alpha_erosion = 1.5
		# Grainsize
		self.D = 4e-3 
		# Critical shear strass
		self.tau_c = 4
		# transport length
		self.transport_length = 4
		# water viscosity
		self.viscosity = 15.e-6

		self.update_morpho_at_input_points = False

		self.threshold_Qw_ratio = 0.


PARAMMORPHO = MorphoParams()
PARAMHYDRO = hydrofunc.PARAMHYDRO



@ti.func
def ZPsi(Z:ti.template(), hw: ti.template(), k_h:ti.template(), k_z:ti.template(), i:ti.i32, j:ti.i32) -> ti.f32:
	'''
	Internal helping function returning the Partitioning surface for sediments (Psi)
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- k_h: the internal coefficient to the shear stress component in the MPM-like entrainment formulation
		- k_z: the internal coefficient to the gravitational component in the MPM-like entrainment formulation
		- i,j: the row col indices
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	rho_ratio = (PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water)/PARAMMORPHO.rho_water
	A = k_z * Z[i,j] + 2 * k_h * hw[i,j]/(PARAMMORPHO.D * rho_ratio)* hsw.Zw(Z,hw,i,j)
	B = k_z + 2 * k_h * hw[i,j]/(PARAMMORPHO.D * rho_ratio)
	return A/B


@ti.func
def SPsi(Z: ti.template(), hw: ti.template(), k_h:ti.template(), k_z:ti.template(), i:ti.template(), j:ti.template(), ir:ti.template(), jr:ti.template())->ti.f32:
	'''
	Internal helping function returning the hydrayulic slope
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- i,j: the row col indices
		- ir,jr: the row col indices of the receivers node
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 20/05/2024)
	'''
	return (ZPsi(Z, hw, k_h, k_z, i,j) - ZPsi(Z, hw, k_h, k_z, ir,jr))/GRID.dx


@ti.kernel
def initiate_step(QsB: ti.template()):
	'''
	Runs the initial operations when running a step
	Mostly a placeholder so far but keeps things consistents
	Arguments:
		- QwB: a 2D field of discharge B (in_temp)
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	# Reinitialise QwB to 0.
	for i,j in QsB:
		QsB[i,j] = 0.


@ti.kernel
def input_discharge_sediment_points(input_rows: ti.template(), input_cols:ti.template(), input_values:ti.template(), QsA: ti.template(), QsB: ti.template(), BCs:ti.template()):
	'''
	Adds discharge in m^3/s into specific input points
	Arguments:
		- input_rows: a 1D field of input point row coordinates (integrer 32 bits)
		- input_cols: a 1D field of input point column coordinates (integrer 32 bits)
		- input_values: a 1D field of input discharges in m^3/s (floating point)
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in_temp)
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''

	for i in input_rows:
		QsA[input_rows[i],input_cols[i]] += input_values[i]
		QsB[input_rows[i],input_cols[i]] += input_values[i]

@ti.kernel
def _compute_Qs_fbal(Z:ti.template(), hw:ti.template(), QsA:ti.template(), QsB:ti.template(), QsC:ti.template(), QwA:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	DENSITY_R = (PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water) * PARAMMORPHO.GRAVITY
	K_EROSION = (PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water)/PARAMMORPHO.rho_water * PARAMMORPHO.GRAVITY/PARAMMORPHO.viscosity * ti.math.pow(PARAMMORPHO.D,2)


	# Traversing each nodes
	for i,j in Z:

		# Case of normal, internal node
		if(gridfuncs.can_out(i,j,BCs) == False):

			# I'll store the hydraulic slopes in this vector
			SPsis = ti.math.vec4(0.,0.,0.,0.)

			# I'll need the sum of the hydraulic slopes in the positive directions
			sumSpsi = 0.

			# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
			## Hydraulic slope
			SSwx = 0.
			SSwy = 0.

			## Topographic slope
			SSZx = 0.
			SSZy = 0.

			## Psi slope
			SSPsix = 0.
			SSPsiy = 0.


			thw = 0.
			
			# Traversing Neighbours
			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1): 
					continue

				if(gridfuncs.can_receive(ir,jr,BCs) == False):
					continue

				# Local topographic slope
				tS = ti.max(hsw.Sw(Z,hw,i,j,ir,jr),0.)

				# Local hydraulic slope
				tSz = ti.max(hsw.Sz(Z,i,j,ir,jr), 0.)

				# Local Partitionning slope
				tSPsi = ti.max(SPsi(Z, hw, PARAMMORPHO.k_h, PARAMMORPHO.k_z, i, j, ir, jr), 0.)

				# # If < 0, neighbour is a donor and I am not interested
				# if(tS <= 0):
				# 	continue

				thw = ti.math.max(thw, ti.max(hsw.Zw(Z,hw,i,j),hsw.Zw(Z,hw,ir,jr)) - ti.max(Z[i,j],Z[ir,jr]))

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSwy):
						SSwy = tS
					if(tSz > SSZy):
						SSZy = tSz
					if(tSPsi > SSPsiy):
						SSPsiy = tSPsi
				else:
					if(tS > SSwx):
						SSwx = tS
					if(tSz > SSZx):
						SSZx = tSz
					if(tSPsi > SSPsix):
						SSPsix = tSPsi

				# Registering local partitioning slope
				SPsis[k] = tSPsi
				# Summing it to global
				sumSpsi += tSPsi

			# Done with processing this particular neighbour

			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			if(sumSpsi == 0.):
				QsC[i,j] = -1
				continue


			# Calculating local norms for the gradients
			gradSw = ti.math.sqrt(SSwx*SSwx + SSwy*SSwy)
			gradSPsi = ti.math.sqrt(SSPsix*SSPsix + SSPsiy*SSPsiy)
			gradSz = ti.math.sqrt(SSZx*SSZx + SSZy*SSZy)

			# Not sure I still need that
			if(gradSw <= 0 or gradSPsi == 0):
				continue

			if(PARAMHYDRO.use_heffmax == False):
				thw = ti.math.max(0., hw[i,j])

			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * K_EROSION * 
				ti.pow( PARAMMORPHO.k_z * gradSz + 
					2 * PARAMMORPHO.k_h * PARAMMORPHO.rho_water/(PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water) 
					/ PARAMMORPHO.D * thw * gradSw - PARAMMORPHO.tau_c,
				PARAMMORPHO.alpha_erosion),
			0.)
			# print(gradSPsi)
			# if(local_erosion_rate > 0):
			# 	print(local_erosion_rate)

			CA = GRID.dx*GRID.dx
			beta = gradSPsi/(PARAMMORPHO.transport_length * sumSpsi)

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QsC[i,j] = (QsA[i,j] + local_erosion_rate * CA)/(1 + CA * beta)

			# If the node cannot give and can only receive, I pass this node
			if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.can_receive(i,j,BCs) == False):
				QsC[i,j] = -1
			else:
				# Transferring flow to neighbours
				for k in range(4):

					# local neighbours
					ir,jr = gridfuncs.neighbours(i,j,k, BCs)
					
					# checking if neighbours
					if(ir == -1):
						continue
					
					# Transferring prop to the hydraulic slope
					ti.atomic_add(QsB[ir,jr], SPsis[k]/sumSpsi * QsC[i,j])
		else:

			gradSz = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val

			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * K_EROSION * ti.pow( gradSz - PARAMMORPHO.tau_c, PARAMMORPHO.alpha_erosion),0.)

			CA = GRID.dx*GRID.dx

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			# QsC[i,j] = (QsA[i,j] + local_erosion_rate * CA)/(1 + CA * PARAMMORPHO.transport_length / GRID.dx)
			QsC[i,j] = QsA[i,j]

	for i,j in Z:
		# Updating local discharge to new time step
		# QsA[i,j] = QsB[i,j]

		adamantic = (PARAMMORPHO.threshold_Qw_ratio != 0)
		if(adamantic):
			if(QwA[i,j] > 0):
				if(abs(1 - QwC[i,j]/QwA[i,j]) < PARAMMORPHO.threshold_Qw_ratio ):
					adamantic = False

		if(QsC[i,j] < 0 or adamantic):
			QsC[i,j] = QsA[i,j]

@ti.kernel
def _compute_Qs_fablink(Z:ti.template(), hw:ti.template(), QsA:ti.template(), QsB:ti.template(), QsC:ti.template(), QwA:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	DENSITY_R = (PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water) * PARAMMORPHO.GRAVITY
	K_EROSION = (PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water)/PARAMMORPHO.rho_water * PARAMMORPHO.GRAVITY/PARAMMORPHO.viscosity * ti.math.pow(PARAMMORPHO.D,2)


	# Traversing each nodes
	for i,j in Z:

		# Case of normal, internal node
		if(gridfuncs.can_out(i,j,BCs) == False):

			# I'll store the hydraulic slopes in this vector
			SPsis = ti.math.vec4(0.,0.,0.,0.)
			heffs = ti.math.vec4(0.,0.,0.,0.)

			# I'll need the sum of the hydraulic slopes in the positive directions
			sumSpsi = 0.

			# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
			## Hydraulic slope
			SSwx = 0.
			SSwy = 0.

			## Topographic slope
			SSZx = 0.
			SSZy = 0.

			## Psi slope
			SSPsix = 0.
			SSPsiy = 0.

			
			# Traversing Neighbours
			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1): 
					continue

				if(gridfuncs.can_receive(ir,jr,BCs) == False):
					continue

				# Local topographic slope
				tS = ti.max(hsw.Sw(Z,hw,i,j,ir,jr),0.)

				# Local hydraulic slope
				tSz = ti.max(hsw.Sz(Z,i,j,ir,jr), 0.)

				# Local Partitionning slope
				tSPsi = ti.max(SPsi(Z, hw, PARAMMORPHO.k_h, PARAMMORPHO.k_z, i, j, ir, jr), 0.)

				heffs[k] = ti.max(hsw.Zw(Z,hw,i,j),hsw.Zw(Z,hw,ir,jr)) - ti.max(Z[i,j],Z[ir,jr])

				# # If < 0, neighbour is a donor and I am not interested
				# if(tS <= 0):
				# 	continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSwy):
						SSwy = tS
					if(tSz > SSZy):
						SSZy = tSz
					if(tSPsi > SSPsiy):
						SSPsiy = tSPsi
				else:
					if(tS > SSwx):
						SSwx = tS
					if(tSz > SSZx):
						SSZx = tSz
					if(tSPsi > SSPsix):
						SSPsix = tSPsi

				# Registering local partitioning slope
				SPsis[k] = tSPsi
				# Summing it to global
				sumSpsi += tSPsi

			# Done with processing this particular neighbour

			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			if(sumSpsi == 0.):
				QsC[i,j] = -1
				continue


			# Calculating local norms for the gradients
			gradSw = ti.math.sqrt(SSwx*SSwx + SSwy*SSwy)
			gradSPsi = ti.math.sqrt(SSPsix*SSPsix + SSPsiy*SSPsiy)
			gradSz = ti.math.sqrt(SSZx*SSZx + SSZy*SSZy)

			# Not sure I still need that
			if(gradSw <= 0 or gradSPsi == 0):
				continue

			local_erosion_rate = 0.
			for k in range(4):
				local_erosion_rate += heffs[k] * SPsis[k]/sumSpsi


			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * K_EROSION * 
				ti.pow( PARAMMORPHO.k_z * gradSz + 
					2. * PARAMMORPHO.k_h * PARAMMORPHO.rho_water/(PARAMMORPHO.rho_sediment - PARAMMORPHO.rho_water) 
					/ PARAMMORPHO.D * local_erosion_rate * gradSw - PARAMMORPHO.tau_c,
				PARAMMORPHO.alpha_erosion),
			0.)
			# print(gradSPsi)
			# if(local_erosion_rate > 0):
			# 	print(local_erosion_rate)

			CA = GRID.dx*GRID.dx
			beta = gradSPsi/(PARAMMORPHO.transport_length * sumSpsi)

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QsC[i,j] = (QsA[i,j] + local_erosion_rate * CA)/(1 + CA * beta)

			# If the node cannot give and can only receive, I pass this node
			if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.can_receive(i,j,BCs) == False):
				QsC[i,j] = -1
			else:
				# Transferring flow to neighbours
				for k in range(4):

					# local neighbours
					ir,jr = gridfuncs.neighbours(i,j,k, BCs)
					
					# checking if neighbours
					if(ir == -1):
						continue
					
					# Transferring prop to the hydraulic slope
					ti.atomic_add(QsB[ir,jr], SPsis[k]/sumSpsi * QsC[i,j])
		else:

			gradSz = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val

			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * K_EROSION * ti.pow( gradSz - PARAMMORPHO.tau_c, PARAMMORPHO.alpha_erosion),0.)

			CA = GRID.dx*GRID.dx

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			# QsC[i,j] = (QsA[i,j] + local_erosion_rate * CA)/(1 + CA * PARAMMORPHO.transport_length / GRID.dx)
			QsC[i,j] = QsA[i,j]

	for i,j in Z:
		# Updating local discharge to new time step
		# QsA[i,j] = QsB[i,j]

		adamantic = (PARAMMORPHO.threshold_Qw_ratio != 0)
		if(adamantic):
			if(QwA[i,j] > 0):
				if(abs(1 - QwC[i,j]/QwA[i,j]) < PARAMMORPHO.threshold_Qw_ratio ):
					adamantic = False

		if(QsC[i,j] < 0 or adamantic):
			QsC[i,j] = QsA[i,j]


@ti.kernel
def _compute_Qs_MPM_like(Z:ti.template(), hw:ti.template(), QsA:ti.template(), QsB:ti.template(), QsC:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# Case of normal, internal node
		if(gridfuncs.can_out(i,j,BCs) == False):

			# I'll store the hydraulic slopes in this vector
			Sws = ti.math.vec4(0.,0.,0.,0.)

			# I'll need the sum of the hydraulic slopes in the positive directions
			sumSlopes = 0.

			# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
			## Hydraulic slope
			SSwx = 0.
			SSwy = 0.
			
			# Traversing Neighbours
			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				# Local topographic slope
				tSw = ti.max(hsw.Sw(Z,hw,i,j,ir,jr),0.)

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tSw > SSwy):
						SSwy = tSw
				else:
					if(tSw > SSwx):
						SSwx = tSw

				# Registering local partitioning slope
				Sws[k] = tSw
				# Summing it to global
				sumSlopes += tSw

			# Done with processing this particular neighbour

			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			if(sumSlopes == 0.):
				QsC[i,j] = -1
				continue


			# Calculating local norms for the gradients
			gradSw = ti.math.sqrt(SSwx*SSwx + SSwy*SSwy)

			# Not sure I still need that
			if(gradSw <= 0):
				continue

			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * ti.math.pow(PARAMMORPHO.rho_water * PARAMMORPHO.GRAVITY * max(hw[i,j],0.) * gradSw - PARAMMORPHO.tau_c, PARAMMORPHO.alpha_erosion), 0.)

			CA = GRID.dx*GRID.dx

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			tK = (1. / PARAMMORPHO.transport_length);
			edotpsy = (local_erosion_rate) / tK;
			C1 = QsA[i,j] / GRID.dx - edotpsy;
			QsC[i,j] = GRID.dx * (edotpsy + C1 * ti.math.exp(-GRID.dx * tK));

			# If the node cannot give and can only receive, I pass this node
			if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.can_receive(i,j,BCs) == False):
				QsC[i,j] = -1
			else:
				# Transferring flow to neighbours
				for k in range(4):

					# local neighbours
					ir,jr = gridfuncs.neighbours(i,j,k, BCs)
					
					# checking if neighbours
					if(ir == -1):
						continue
					
					# Transferring prop to the hydraulic slope
					ti.atomic_add(QsB[ir,jr], Sws[k]/sumSlopes * QsC[i,j])
		else:

			gradSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val

			local_erosion_rate = ti.max(PARAMMORPHO.k_erosion * ti.math.pow(PARAMMORPHO.rho_water * PARAMMORPHO.GRAVITY * max(hw[i,j],0.) * gradSw - PARAMMORPHO.tau_c, PARAMMORPHO.alpha_erosion), 0.)

			CA = GRID.dx*GRID.dx

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			# QsC[i,j] = (QsA[i,j] + local_erosion_rate * CA)/(1 + CA * PARAMMORPHO.transport_length / GRID.dx)
			QsC[i,j] = QsA[i,j]

	for i,j in Z:
		# Updating local discharge to new time step
		# QsA[i,j] = QsB[i,j]

		if(QsC[i,j] < 0):
			QsC[i,j] = QsA[i,j]
			









@ti.kernel
def _compute_hs(Z:ti.template(), hw:ti.template(), QsA:ti.template(), QsB:ti.template(), QsC:ti.template(), BCs:ti.template() ):
	'''
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	# Traversing nodes
	for i,j in Z:

		if(gridfuncs.can_receive(i,j,BCs) == False and PARAMMORPHO.update_morpho_at_input_points == False):
			continue
		
		# Only where nodes are active (i.e. flow cannot leave and can traverse)
		# if(gridfuncs.is_active(i,j,BCs) == False):
		# 	continue

		dz = (QsA[i,j] - QsC[i,j]) * PARAMMORPHO.dt_morpho/(GRID.dx*GRID.dy)

		# Updating flow depth (cannot be < 0)
		Z[i,j] += dz
		hw[i,j] = hw[i,j] - dz
		QsA[i,j] = QsB[i,j]




########################################################################
########################################################################
############### EXPOSED API ############################################
########################################################################
########################################################################

# Note that the input functions are already exposed

compute_hs = None
compute_Qs = None


def set_morpho_CC():
	'''
	Expose the right API function of the comnpile time parameters in PARAMHYDRO.
	Need to be called after setting hte different parameters in the singleton PARAMHYDRO
	Returns:
		- Nothing, update global function refs
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	


	# # fectch neighbours placeholder
	global compute_hs
	global compute_Qs

	
	# Feed it
	if(PARAMMORPHO.morphomode == MorphoMode.fbal):
		compute_hs = _compute_hs
		compute_Qs = _compute_Qs_fbal
	elif(PARAMMORPHO.morphomode == MorphoMode.MPM_like):
		compute_hs = _compute_hs
		compute_Qs = _compute_Qs_MPM_like
	elif(PARAMMORPHO.morphomode == MorphoMode.fbal_links):
		compute_hs = _compute_hs
		compute_Qs = _compute_Qs_fablink
	else:
		raise NotImplementedError('PARAMMORPHO Not implemented yet')



'''
Riverdale environment:

This is the main high level class interfacing with the model. 
This is the API users use to get the data out of the model and run time steps.
The other high level class class managing the model inputs and parameters is the RDParams class (see rd_params.py) 
The (internal) structure is a bit convoluted for the sake of optimising calculation time.

'''

import taichi as ti
import numpy as np
from enum import Enum
import dagger as dag
import scabbard.riverdale.rd_params as rdpa
import scabbard.riverdale.rd_grid as rdgd
import scabbard.riverdale.rd_flow as rdfl
import scabbard.riverdale.rd_hydrodynamics as rdhy
import scabbard.riverdale.rd_morphodynamics as rdmo
import scabbard.riverdale.rd_LM as rdlm
import scabbard as scb
from scipy.ndimage import gaussian_filter


# @ti.data_oriented
class Riverdale:
	'''
	Main class controlling riverdale's model.
	Cannot be directly instantiated, please use the factory function at the end of this script.
	'''

	#
	# _already_created = False
	# _instance_created = False
	
	def __init__(self):

# 		if not Riverdale._instance_created:
# 			raise Exception("Riverdale cannot be instantiated directly. Please use the factory functions.")

# 		if Riverdale._already_created:
# 			raise Exception("""
# Riverdale cannot be instantiated more than once so far within the same runtime (~= within the same script). 
# This is because only one unique taichi lang context can exist at once as far as I know, I am working on solutions 
# but in the meantime you can run the model in batch using subprocess.
# """)
		
		# Place holders for the different variables
		## This is the parameter sheet
		self.param = None

		## Model-side params
		self.GRID = rdgd.GRID
		self.PARAMHYDRO = rdhy.PARAMHYDRO
		self.PARAMMORPHO = rdmo.PARAMMORPHO


		## Fields for hydro
		self.QwA = None
		self.QwB = None
		self.QwC = None
		self.Z = None
		self.hw = None
		self.P = None
		self.input_rows_Qw = None
		self.input_cols_Qw = None
		self.input_Qw = None
		self.convrat = None
		self.constraints = None

		self.fdir = None
		self.fsurf = None

		## Field for morpho
		self.QsA = None
		self.QsB = None
		self.QsC = None
		self.input_rows_Qs = None
		self.input_cols_Qs = None
		self.input_Qs = None

		# temporary fields
		self.temp_fields = {ti.u8:[], ti.i32:[], ti.f32:[], ti.f64:[]}



	def run_hydro(self, n_steps, recompute_flow = False, expe_N_prop = 0, expe_CFL_variable = False, flush_LM = False):
		'''
		Main runner function for the hydrodynamics part of the model.
		NOte that all the parameters have been compiled prior to running that functions, so not much to control here
		Arguments:
			- nsteps: The number of time step to run
		returns:
			- Nothing, update the model internally
		Authors:
			- B.G (last modification 20/05/2024)
		'''

		if(recompute_flow):
			if(self.param.use_fdir_D8):
				rdfl.compute_D4_Zw(self.Z, self.hw, self.fdir, self.BCs)
			else:
				fsurf = scb.raster.raster_from_array(self.Z.to_numpy()+self.hw.to_numpy(),self.param._dx)
				tBCs = None if self.param.BCs is None else self.BCs.to_numpy()
				scb.flow.priority_flood(fsurf, in_place = True,BCs = tBCs)
				scb.filters.gaussian_fourier(fsurf, in_place = True, magnitude = 50,BCs = tBCs)				 
				scb.flow.priority_flood(fsurf, in_place = True,BCs = tBCs)
				self.fsurf.from_numpy(fsurf.Z.astype(np.float32))

		# Running loop
		for it in range(n_steps):
			
			# Initialising step: setting stuff to 0, reset some counters and things like that
			self._run_init_hydro()
			
			# Add external water inputs: rain, discrete entry points, ...
			self._run_hydro_inputs()

			# Actually runs the runoff simulation
			# self._run_hydro() if expe_N_prop <= 1 or (it % expe_N_prop == 0) else rdhy._propagate_QwA_only(self.Z, self.hw, self.QwA, self.QwB, self.BCs )
			self._run_hydro(expe_CFL_variable = expe_CFL_variable, flush_LM = flush_LM)

		# That's it reallly, see bellow for the internal functions

	def _run_init_hydro(self):
		rdhy.initiate_step(self.QwB)

	def _run_hydro_inputs(self):
		'''
		Internal function automating the runniong of anything that adds water to the model (precipitations, input points, ...)
		It determines everything automatically from the param sheet (RDparam class)
		
		Returns: 
			- Nothing, runs a subpart of the model
		
		Authors:
			- B.G (last modification: 28/05/2024)
		'''

		# First checking if we need the  precipitations inputs
		if(self.param.need_precipitations):

			# then running the 2D precipitations
			if(self.param.precipitations_are_2D):
					rdhy.variable_rain(self.QwA, self.QwB, self.P,self.BCs)
			# or the 1D
			else:
				rdhy.constant_rain(self.QwA, self.QwB, self.P,self.BCs)
		# Secondly, applying the discrete inputs if needed too				
		if(self.param.need_input_Qw):
			rdhy.input_discharge_points(self.input_rows_Qw, self.input_cols_Qw, self.input_Qw, self.QwA, self.QwB, self.BCs)

	def _run_hydro(self, expe_CFL_variable = False, flush_LM = False):
		'''
		Internal runner for hydro functions

		'''

		if(flush_LM):
			rdhy._flush_QwA_only(self.Z, self.hw, self.QwA, self.QwB, self.BCs, self.fdir)
			return

		if(self.param.use_fdir_D8):
			if(self.param.stationary):
				rdhy._compute_Qw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs, self.fdir) #if rdhy.FlowMode.static_drape != self.param._hydro_compute_mode else rdhy.compute_Qw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.constraints, self.BCs)
			else:
				rdhy._compute_Qw_dynamic(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs) #if rdhy.FlowMode.static_drape != self.param._hydro_compute_mode else rdhy.compute_Qw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.constraints, self.BCs)
		else:
			rdhy._compute_Qw_surfrec(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs, self.fsurf) #if rdhy.FlowMode.static_drape != self.param._hydro_compute_mode else rdhy.compute_Qw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.constraints, self.BCs)
		
		if(expe_CFL_variable):
			rdhy._compute_hw_CFL(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs, 1e-4, 0.001 )
		else:	
			if(self.param.stationary):		
				rdhy.compute_hw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs) if rdhy.FlowMode.static_drape != self.param.hydro_compute_mode else rdhy.compute_hw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.constraints, self.BCs)
			else:
				rdhy.compute_hw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.BCs) if rdhy.FlowMode.static_drape != self.param.hydro_compute_mode else rdhy.compute_hw(self.Z, self.hw, self.QwA, self.QwB, self.QwC, self.constraints, self.BCs)
				
	def raise_analytical_hw(self):
		'''
		This function uses the current fields of hw and Qwin (QwA) to calculate an analytical hw.
		This is useful in multiple occasions, for example speed up convergence of hillslopes if the rest is fine or creating an initial guess from an external Qwin.
		WARNING: this is not a magical analytical solution, most of the time it creates a noisy combination of flow depth with walls. 
		If 2D SWEs had a true easy analytical solution not involving gigantic inverse matrices it would be known

		Arguments:
			- None
		returns:
			- Nothing, update the model internally
		Authors:
			- B.G (last modification 09/2024)
		'''
		temp, = self.query_temporary_fields(1)
		rdhy._raise_analytical_hw(self.Z, self.hw, self.QwA, temp, self.BCs)

	def diffuse_hw(self, n_steps = 100):

		temp, = self.query_temporary_fields(1)
		for i in range(n_steps):
			rdhy._CA_smooth(self.Z, self.hw, temp, self.BCs)
		ti.sync()
		

	def propagate_QwA(self, SFD = True):

		self._run_init_hydro()
		self._run_hydro_inputs()
		input_values = self.QwB.to_numpy()

		if(SFD):
			tZ = (self.Z.to_numpy() + self.hw.to_numpy())
			tBCs = scb.flow.get_normal_BCs(tZ) if self.param.BCs is None else self.BCs.to_numpy()
			stg = scb.flow.SFGraph(tZ, BCs = tBCs, D4 = True, dx = self.param._dx, backend = 'ttb', fill_LM = True, step_fill = 1e-3)
			Qw = scb.flow.propagate(stg, input_values, step_fill = 1e-3)
		else:
			tZ = (self.Z.to_numpy() + self.hw.to_numpy())
			tBCs = scb.flow.get_normal_BCs(tZ) if self.param.BCs is None else self.BCs.to_numpy()
			grid = scb.raster.raster_from_array(tZ, dx = self.param._dx, xmin = 0., ymin = 0., dtype = np.float32)
			Qw = scb.flow.propagate(grid, input_values, method = 'mfd_S', BCs = tBCs, D4 = True, fill_LM = True, step_fill = 1e-3)

		self.QwA.from_numpy(Qw.astype(np.float32))



	def run_morpho(self, n_steps):
		'''
		Main runner function for the morphodynamics part of the model.
		NOte that all the parameters have been compiled prior to running that functions, so not much to control here
		Arguments:
			- nsteps: The number of time step to run
		returns:
			- Nothing, update the model internally
		Authors:
			- B.G (last modification 20/05/2024)
		'''
		for _ in range(n_steps):
			###
			# TODO:: Add and test morpho stuff here
			self._run_init_morpho()

			# NOT READY YET
			# self._run_inputs_Qs()

			self._run_morpho()



	

	def _run_init_morpho(self):
		rdmo.initiate_step(self.QsB)

	def _run_inputs_Qs(self):
		rdmo.input_discharge_sediment_points(self.input_rows_Qs, self.input_cols_Qs, self.input_Qs, self.QsA, self.QsB, self.BCs)

	def _run_morpho(self):
		rdmo.compute_Qs(self.Z, self.hw, self.QsA, self.QsB, self.QsC, self.QwA, self.QwC, self.BCs )
		rdmo.compute_hs(self.Z, self.hw, self.QsA, self.QsB, self.QsC, self.BCs )


	@property
	def convergence_ratio(self):

		if(self.param is None):
			raise ValueError('cannot return convergence ratio if the model is not initialised')

		if(self.convrat is None):
			self.convrat = ti.field(dtype = self.param.dtype_float, shape = ())

		rdhy.check_convergence(self.QwA, self.QwC, 0.01, self.convrat, self.BCs)

		return float(self.convrat.to_numpy())

	def get_GridCPP(self):
		'''
		PENDING DEPRECATION
		Returns a GridCPP object corresponding to the grid geometry and boundary conditions.
		GridCPP objectss are used to interact with the DAGGER c++ engine which I use for CPU intensive tasks.
		It will probably also be used for communication with TTBlib and fastscapelib

		Returns:
			- a GridCPP object ready to be passed to the C++ engine
		Authors:
			- B.G. (last modification: 30/05/2024)
		'''
		return dag.GridCPP_f32(self.param._nx,self.param._ny,self.param._dx,self.param._dy,0 if self.param.boundaries == rdgd.BoundaryConditions.normal else 3)


	def query_temporary_fields(self, N, dtype = 'f32'):
		'''
		Ask riverdale for a number of temporary fields of a given type. 
		Effectively avoids unecessary new fields and memory leaks.
		It only creates a temporary field if it does not exist yet.
			
		Arguments:
			- N: the number of fields to return
			- dtype: the data type: f32, u8 or i32 so far
		Returns:
			- a tuple with all the fields
		Authors:
			- B.G. (last modification: 12/06/2024)
		'''
		
		# Data type conversion to the dict key
		if(dtype in ['f32',self.param.dtype_float, np.float32]):
			dtype = ti.f32
		elif(dtype in ['u8',np.uint8, ti.u8]):
			dtype = ti.u8
		elif(dtype in ['i32',np.int32, ti.i32]):
			dtype = ti.i32
		else:
			raise TypeError(f"dtype {dtype} is not recognised. Should be one of ['f32','u8','i32'] or their taichi/numpy equivalents (e.g. np.float32 or ti.f32)")

		# gathering the results in a list
		output = []
		for i in range(N):
			# DO I need to create the field or does it exist already?
			if(i >= len(self.temp_fields[dtype])):
				self.temp_fields[dtype].append(ti.field(dtype=dtype, shape = (self.GRID.ny,self.GRID.nx)))
			# Filling it with 0s by default	
			self.temp_fields[dtype][i].fill(0)
			#saving a ref in the list
			output.append(self.temp_fields[dtype][i])

		# Done, returning the list conerted into a tuple
		return tuple(output)



	def save(self, fname = 'riverdale_run'):
		'''
		Experiments on saving files
		'''

		import pickle
		# import copy
		param = self.param
		param._RD = None
		tosave = {'param':param, 'hw':self.hw.to_numpy(), 'Z':self.Z.to_numpy(), 'QwA':self.QwA.to_numpy(), 'QwC':self.QwC.to_numpy()}
		with open(fname+'.rvd', 'wb') as f:
			pickle.dump(tosave, f)








# External factory function
def create_from_params(param):
	'''
		This function creates the instance of the model and binds the input param sheet
		In some sort it "compiles" the model. 
		IMPORTANT: every static variable and constants will be fixed at that point
	'''

	# Generating the empty instance, which should be the sole one
	# instance = Riverdale._create_instance()
	instance = Riverdale()

	# Referencing it in the Param files thus fixing it 
	param._RD = instance
	instance.param = param

	if(
		param._boundaries != rdgd.BoundaryConditions.normal 
		and param._boundaries != rdgd.BoundaryConditions.customs
	):
		raise Exception("Selected Boundary conditions is not available yet (WIP)")

	# Setting up the grid
	instance.GRID.dx = param._dx
	instance.GRID.dy = param._dy
	instance.GRID.nx = param._nx
	instance.GRID.ny = param._ny
	instance.GRID.nxy = param._nxy
	instance.GRID.boundaries = param._boundaries

	instance.PARAMHYDRO.hydro_slope_bc_mode = int(param._boundary_slope_mode.value)
	instance.PARAMHYDRO.hydro_slope_bc_val = param._boundary_slope_value
	instance.PARAMHYDRO.flowmode = param.hydro_compute_mode
	instance.PARAMHYDRO.clamp_div_hw_val = param._clamp_div_hw_val
	# instance.PARAMHYDRO.flowmode = param.hydro_compute_mode


	if(param.BCs is None):
		instance.BCs = ti.field(ti.int32, shape = (1,1))
	else:
		instance.BCs = ti.field(ti.uint8, shape = (param._ny, param._nx))
		instance.BCs.from_numpy(param.BCs)

	# Compiling the grid functions
	rdgd.set_grid_CC()

	# Setting up the flow conditions
	instance.PARAMHYDRO.manning = param.manning
	instance.PARAMHYDRO.dt_hydro = param.dt_hydro

	#WILL NEED TO ADD THE OPTION  LATER
	instance.fdir = ti.field(ti.u8, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.PARAMMORPHO.use_original_dir_for_LM = True
	instance.PARAMHYDRO.use_original_dir_for_LM = True
	instance.PARAMHYDRO.LM_pathforcer = param._LM_npath
	if(param.use_fdir_D8):
		instance.fsurf = ti.field(ti.f32, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.PARAMMORPHO.use_original_dir_for_LM = False
	instance.PARAMHYDRO.use_original_dir_for_LM = False

	

	# Compiling the hydrodynamics
	rdhy.set_hydro_CC()

	instance.QwA = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.QwA.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))
	instance.QwB = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.QwB.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))
	instance.QwC = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.QwC.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))
	instance.constraints = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx,2)) if rdhy.FlowMode.static_drape == instance.param._hydro_compute_mode else None

	instance.Z = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
	instance.Z.from_numpy(param.initial_Z)
	instance.hw = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
	
	if(param.initial_hw is not None):
		instance.hw.from_numpy(param.initial_hw)
	else:
		instance.hw.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))

	if param.precipitations_are_2D:
		instance.P = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx)) 
		instance.P.from_numpy(param.precipitations.astype(np.float32))
	else:
		instance.P = param.precipitations

	if(param.need_input_Qw):
		n_inputs_QW = param._input_rows_Qw.shape[0]
		instance.input_rows_Qw = ti.field(ti.i32, shape = (n_inputs_QW))
		instance.input_rows_Qw.from_numpy(param._input_rows_Qw)
		instance.input_cols_Qw = ti.field(ti.i32, shape = (n_inputs_QW))
		instance.input_cols_Qw.from_numpy(param._input_cols_Qw)
		instance.input_Qw = ti.field(instance.param.dtype_float, shape = (n_inputs_QW))
		instance.input_Qw.from_numpy(param._input_Qw)

	if(param.precompute_Qw):
		# Precomputing a D8 propagation of QwA within the cpu
		instance.propagate_QwA()


	if(param.morpho):

		instance.PARAMMORPHO.dt_morpho = param.dt_morpho
		instance.PARAMMORPHO.morphomode = param.morphomode
		instance.PARAMMORPHO.GRAVITY = param.GRAVITY
		instance.PARAMMORPHO.rho_water = param.rho_water
		instance.PARAMMORPHO.rho_sediment = param.rho_sediment
		instance.PARAMMORPHO.k_z = param.k_z
		instance.PARAMMORPHO.k_h = param.k_h
		instance.PARAMMORPHO.k_erosion = param.k_erosion
		instance.PARAMMORPHO.alpha_erosion = param.alpha_erosion
		instance.PARAMMORPHO.D = param.D
		instance.PARAMMORPHO.tau_c = param.tau_c
		instance.PARAMMORPHO.transport_length = param.transport_length

		rdmo.set_morpho_CC()

		instance.QsA = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
		instance.QsA.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))
		instance.QsB = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
		instance.QsB.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))
		instance.QsC = ti.field(instance.param.dtype_float, shape = (instance.GRID.ny,instance.GRID.nx))
		instance.QsC.from_numpy(np.zeros((param._ny,param._nx), dtype = np.float32))

	
	instance.input_rows_Qs = None
	instance.input_cols_Qs = None
	instance.input_Qs = None


	#running eventual preprocessors
	if(param.use_fdir_D8):
		# print('debug info: computing fdir')
		# creating the original hydraulic pattern by pre filling the topography with water
		rdlm.priority_flood(instance)
		# Calculating the motherflow direction, used to trasfer Qw out of local minimas
		rdfl.compute_D4_Zw(instance.Z, instance.hw, instance.fdir, instance.BCs)
		# print(np.unique(instance.fdir.to_numpy()))
	else:
		fsurf = scb.raster.raster_from_array(instance.Z.to_numpy()+instance.hw.to_numpy(),instance.param._dx)
		tBCs = None if instance.param.BCs is None else instance.BCs.to_numpy()
		scb.flow.priority_flood(fsurf, in_place = True,BCs = tBCs)
		scb.filters.gaussian_fourier(fsurf, in_place = True, magnitude = 50,BCs = tBCs)				 
		scb.flow.priority_flood(fsurf, in_place = True,BCs = tBCs)
		instance.fsurf.from_numpy(fsurf.Z.astype(np.float32))

	return instance


def load_riverdale(fname):
	import pickle
	with open(fname, 'rb') as f:
		loaded_dict = pickle.load(f)

	rd = create_from_params(loaded_dict['param'])

	rd.hw.from_numpy(loaded_dict['hw'])
	rd.Z.from_numpy(loaded_dict['Z'])
	rd.QwA.from_numpy(loaded_dict['QwA'])
	rd.QwC.from_numpy(loaded_dict['QwC'])
	
	return rd
































































# end of file



import h5py
import numpy as np

def save_array_in_group(file_path, array, group_name, dataset_name, overwrite=True):
	"""Save an array within a specified group in an HDF5 file, with optional overwrite."""
	with h5py.File(file_path, 'a') as f:
		group = f.require_group(group_name)  # Create the group if it doesn't exist
		
		if dataset_name in group:
			if overwrite:
				del group[dataset_name]  # Delete existing dataset if overwrite is enabled
			else:
				print(f"Dataset '{dataset_name}' already exists in '{group_name}'. Skipping.")
				return
		
		group.create_dataset(dataset_name, data=array, compression="gzip")

def load_array_from_group(file_path, group_name, dataset_name):
	"""Load an array from a specified group (subfolder) in an HDF5 file."""
	with h5py.File(file_path, 'r') as f:
		return f[f"{group_name}/{dataset_name}"][()]

def load_all_arrays_in_group(file_path, group_name):
	"""Load all arrays within a specific group into a dictionary."""
	arrays = {}
	with h5py.File(file_path, 'r') as f:
		group = f[group_name]
		for name in group:
			arrays[name] = group[name][()]
	return arrays


def inspect_hdf5(file_path):
	"""Inspect and print details of an HDF5 file."""
	with h5py.File(file_path, 'r') as f:
		def explore(name, obj):
			if isinstance(obj, h5py.Dataset):
				print(f"Dataset: {name}")
				print(f" - Shape: {obj.shape}")
				print(f" - Data Type: {obj.dtype}")
				if obj.compression:
					print(f" - Compression: {obj.compression}")
					print(f" - Compression Options: {obj.compression_opts}")
				else:
					print(" - Compression: None")
			elif isinstance(obj, h5py.Group):
				print(f"Group: {name}")

		# Walk through each item in the HDF5 file
		f.visititems(explore)


def list_datasets_in_group(file_path, group_name):
	"""Return a list of all dataset names in a specified group."""
	with h5py.File(file_path, 'r') as f:
		if group_name in f:
			group = f[group_name]
			return [name for name in group if isinstance(group[name], h5py.Dataset)]
		else:
			print(f"Group '{group_name}' does not exist in the file.")
			return []


'''
This files contains the functions and compile-time constant to naviagate and test the grid from the GPU point of views. That includes:
- Neighbouring operations
- Management of the boundary conditions
- Grid dimensions and other geometrical characteristics


Riverdale's convention for the nth neighbouring:

D4:

|   | 0 |   |
| 1 |i,j| 2 |
|   | 3 |   |

In other words:
- k=0: top neighbour
- k=1: left neighbour
- k=2: right neighbour
- k=3: bottom neighbour

B.G - 26/04/1992 - Acigné

'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 


class BoundaryConditions(Enum):
	'''
	Enumeration of the different boundary condition types possible
	'''	
	normal = 0
	periodicEW = 1
	periodicNS = 2
	customs = 3

class BoundaryConditionsSlope(Enum):
	'''
	Enumeration of the different boundary condition types possible
	'''	
	fixed_elevation = 0
	fixed_slope = 1



@scaut.singleton
class GridParams:
	'''
		Holds all the constants describing the grid dimensions and boundary conditions
		Needs to be set before compiling the taichi kernels and funcs and cannot be changed dynamically
		This part of the code might evolve later, the reason behind this structure is to force these element to be
		compile-time constants as it boosts performances (we are talking 10 - 20% speed up compared to having them as parameters of functions)
	'''
	def __init__(self):
		# Grid parameters
		self.dx = 1.
		self.dy = 1.
		self.nx = 1.
		self.ny = 1.
		self.nxy = 1.
		self.boundaries = BoundaryConditions.normal





# Class Holder
GRID = GridParams()



#################################################################################################
############################## Normal Boundaries ################################################
#################################################################################################

# These boundaries represents the classic DEM boundaries where all the nodes are valid but flow can escape from the boundaries



@ti.func
def _check_top_row_normal(i:int, j:int, k:int, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	
	# Only checking if it actually is on the first row
	if(i == 0):
		# Checking all the different cases: firs, last cols and the middle
		if((j == 0 and k <= 1) or (j == GRID.nx-1 and (k == 0 or k == 2)) or (k==0)):
			valid = False
	# Done
	return valid

@ti.func
def _check_leftest_col_normal(i:int, j:int, k:int, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	
	# Only checking if it actually is on the first col
	if(j == 0):
		if(k==1):
			valid = False
	# Done
	return valid

@ti.func
def _check_rightest_col_normal(i:int, j:int, k:int, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	
	# Only checking if it actually is on the first col
	if(j == GRID.nx-1):
		if(k==2):
			valid = False
	# Done
	return valid

@ti.func
def _check_bottom_row_normal(i:int, j:int, k:int, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	
	# Only checking if it actually is on the first row
	if(i == GRID.ny-1):
		# Checking all the different cases: firs, last cols and the middle
		if((j == 0 and (k == 1 or k == 3)) or (j == GRID.nx-1 and (k == 3 or k == 2)) or (k==3)):
			valid = False
	# Done
	return valid

@ti.func
def _cast_neighbour_normal(i:int, j:int, k:int, valid:bool):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 30/04/2024)
	'''

	# Preformat the output
	ir,jr = -1,-1

	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir,jr = i-1, j
		if(k == 1):
			ir,jr = i, j-1
		if(k == 2):
			ir,jr = i, j+1
		if(k == 3):
			ir,jr = i+1, j

	return ir, jr

@ti.func
def _neighbours_normal(i:int, j:int, k:int, BCs:ti.template()):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D4) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification on the 29th of April)
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_normal(i,j,k,valid)
	valid = _check_leftest_col_normal(i,j,k,valid)
	valid = _check_rightest_col_normal(i,j,k,valid)
	valid = _check_bottom_row_normal(i,j,k,valid)

	# getting the actual neighbours
	return _cast_neighbour_normal(i,j,k,valid)

@ti.func
def _is_active_normal(i:int, j:int, BCs:ti.template()):
	'''
		Quick utility function determining if a node is active or not for normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True if the node is active
			- False if inactive (i.e in this case outs)
	'''
	valid = True
	# if(i==0 or j == 0 or i == GRID.ny - 1 or j == GRID.nx - 1):
	# 	valid = False
	return valid

@ti.func
def _can_receive_normal(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = True
	return valid

@ti.func
def _can_give_normal(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = True
	if(i==0 or j == 0 or i == GRID.ny - 1 or j == GRID.nx - 1):
		valid = False
	return valid


@ti.func
def _can_out_normal(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = False
	if(i==0 or j == 0 or i == GRID.ny - 1 or j == GRID.nx - 1):
		valid = True
	return valid



#################################################################################################
############################## Customs Boundaries ###############################################
#################################################################################################

'''
Reminder, I am using the DAGGER convention
// Cannot flow at all = nodata
NO_FLOW = 0,

// Internal Node (can flow in every directions)
FLOW = 1,

// Internal Node (can flow in every directions) BUT neighbours a special flow
// condition and may need specific care
FLOW_BUT = 2,

// flow can out there but can also flow to downstream neighbours
CAN_OUT = 3,

// flow can only out from this cell
OUT = 4,

// Not only flow HAS to out there: neighbouring flows will be drained there no
// matter what
FORCE_OUT = 5,

// Flows through the cell is possible, but the cell CANNOT out fluxes from
// this boundary (reserved to model edges, internal boundaries wont give to
// nodata anyway)
CANNOT_OUT = 6,

// Flow can only flow to potential receivers
IN = 7,

// Forced INFLOW: flow will flow to all neighbours (except other FORCE_IN)
FORCE_IN = 8,

// periodic border
PERIODIC_BORDER = 9
'''


@ti.func
def _check_top_row_customs(i:int, j:int, k:int, BCs:ti.template(), valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# I assume it's good
	# Only checking if it actually is on the first row
	if(i == 0):
		# Checking all the different cases: firs, last cols and the middle
		if(k==0):
			valid = False
	# Done
	return valid

@ti.func
def _check_leftest_col_customs(i:int, j:int, k:int, BCs:ti.template(), valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# I assume it's good
	# Only checking if it actually is on the first col
	if(j == 0):
		if(k==1):
			valid = False
	# Done
	return valid

@ti.func
def _check_rightest_col_customs(i:int, j:int, k:int, BCs:ti.template(), valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# I assume it's good
	# Only checking if it actually is on the first col
	if(j == GRID.nx-1):
		if(k==2):
			valid = False
	# Done
	return valid

@ti.func
def _check_bottom_row_customs(i:int, j:int, k:int, BCs:ti.template(), valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# I assume it's good
	# Only checking if it actually is on the first row
	if(i == GRID.ny-1):
		# Checking all the different cases: firs, last cols and the middle
		if(k==3):
			valid = False
	# Done
	return valid

@ti.func
def _cast_neighbour_customs(i:int, j:int, k:int, valid:bool, BCs:ti.template()):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''

	# Preformat the output
	ir,jr = -1,-1

	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir,jr = i-1, j
		if(k == 1):
			ir,jr = i, j-1
		if(k == 2):
			ir,jr = i, j+1
		if(k == 3):
			ir,jr = i+1, j

	if(ir == -1 or jr == -1):
		ir,jr = -1,-1
	elif(BCs[i,j] == 0):
		ir,jr = -1,-1
	elif(BCs[ir,jr] == 0):
		ir,jr = -1,-1
		

	return ir, jr

@ti.func
def _neighbours_customs(i:int, j:int, k:int, BCs:ti.template()):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D4) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification 02/05/2024)
	TODO:
		- adding the Periodic boundary management in the checks
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_customs(i,j,k,BCs,valid)
	valid = _check_leftest_col_customs(i,j,k,BCs,valid)
	valid = _check_rightest_col_customs(i,j,k,BCs,valid)
	valid = _check_bottom_row_customs(i,j,k,BCs,valid)

	# getting the actual neighbours
	return _cast_neighbour_customs(i,j,k,valid,BCs)

@ti.func
def _can_receive_customs(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = True
	if(BCs[i,j] == 6 or BCs[i,j] == 7 or BCs[i,j] == 8 or BCs[i,j] == 0):
		valid = False
	return valid

@ti.func
def _can_give_customs(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i,j] == 1 or BCs[i,j] == 6 or BCs[i,j] == 7 or BCs[i,j] == 8 or BCs[i,j] == 9):
		valid = True
	return valid


@ti.func
def _can_out_customs(i:int, j:int, BCs:ti.template()):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i,j] == 3 or BCs[i,j] == 4 or BCs[i,j] == 5):
		valid = True
	return valid


@ti.func
def _is_active_customs(i:int, j:int, BCs:ti.template()):
	'''
		Quick utility function determining if a node is active or not for normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True if the node is active
			- False if inactive (i.e in this case outs)
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = True
	if(BCs[i,j] == 0): #) or _can_out_customs(i,j,BCs) or _can_receive_customs(i,j,BCs) == False):
		valid = False
	return valid



########################################################################
########################################################################
############### GENERIC  FUNCTIONS #####################################
########################################################################
########################################################################


@ti.func
def oppk(k:ti.i32):
	'''
	Returns the opposite neighbour code, e.g. if supplied with 1 (left neighbour), returns 2 (right neighbour)
	Useful to check if a k neighbour points toward a cell
	
	Arguments:
		k: the neihgbour code

	returns:
		The opposite neighbour code

	Authors:
		- B.G. (last modifications: 06/2024)
	'''
	return 3 if k == 0 else (2 if k == 1 else (1 if k == 2 else (0 if k == 3 else 5)))




########################################################################
########################################################################
############### EXPOSED API ############################################
########################################################################
########################################################################

# Exporting the generic API
neighbours = None
is_active = None
can_receive = None
can_give = None
can_out = None

def set_grid_CC():
	'''
		Python side to be called before compilation and set the generic functions matching the grid criterion and boundary conditions
	'''
	# fectch neighbours placeholder
	global neighbours
	global is_active
	global can_receive
	global can_give
	global can_out
	
	# Feed it
	if(GRID.boundaries == BoundaryConditions.normal):
		neighbours = _neighbours_normal
		is_active = _is_active_normal
		can_receive = _can_receive_normal
		can_give = _can_give_normal
		can_out = _can_out_normal
	elif(GRID.boundaries == BoundaryConditions.customs):
		# print('DEBUG::BC::CUSTOM')
		neighbours = _neighbours_customs
		is_active = _is_active_customs
		can_receive = _can_receive_customs
		can_give = _can_give_customs
		can_out = _can_out_customs
	else:
		raise NotImplementedError('BC Not implemented yet')



'''
Collections of tools related to kernels
For example Kernel Density Estimator
Spreading a values following a kernel on a numpy array
...
'''

import numpy as np


def sinusoidal_wave(X, a, b, L):
    # Compute Y values for the sinusoidal wave
    Y = (b - a) / 2 * np.sin(2 * np.pi * X / L) + (a + b) / 2
    return Y

def generate_u_shaped_sloped_surface(nx, ny, dx, dy, slope=0.1, Umag = 0.8):
    """
    Generates a 2D numpy array describing a U-shaped surface that is concave (facing upwards)
    and has a slope making the northern part the highest point.
    
    Parameters:
    - nx, ny: Number of cells in the east-west (x) and north-south (y) directions.
    - dx, dy: Cell size in the east-west (x) and north-south (y) directions.
    - slope: Slope of the surface in the north-south direction, making the north higher.
    
    Returns:
    - 2D numpy array representing the surface elevation.
    """
    
    # Generate meshgrid for the positions
    x, y = np.meshgrid(np.arange(nx) * dx, np.arange(ny) * dy)
    
    # Center the x values around 0 for the U shape calculation
    x_centered = x - (nx * dx) / 2
    
    # Calculate the U shape using a negative quadratic function (for concave shape)
    u_shape = -x_centered**2
    
    # Normalize U shape to have values between 0 and 1
    u_shape_normalized = (u_shape - np.min(u_shape)) / (np.max(u_shape) - np.min(u_shape)) * Umag
    
    # Modify the slope effect to make the northern part the highest
    slope_effect = (ny * dy - y) * slope
    
    # Combine the U shape and the slope to generate the surface
    surface = u_shape_normalized*-1 + slope_effect
    
    # Normalize the surface to have a minimum of 0 (optional)
    surface -= np.min(surface)
    
    return surface

def gaussian_spread_on_1D(X = np.linspace(0, 100-1, 100), M = 50, x_c = 50, sigma = 10):

	# Calculate Gaussian
	gaussian = (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-0.5 * ((X - x_c) / sigma) ** 2)

	# Scale the Gaussian so its sum equals M
	gaussian *= M / np.sum(gaussian)

	# Verify the sum of the Gaussian is approximately M
	print("Sum of Gaussian values:", np.sum(gaussian))

	return gaussian



'''
General sets of utilities to check type, characterise functions, classes, ...

B.G.
'''
import numpy as np



def singleton(cls):
	'''
	Decorator ensuring a classes becomes a singleton (= can only be instantiated once)
	
	To use: 
	@singleton
	class myclass:
		...
	
	Then myclass canonly be instantiated once

	Authors:
		- B.G. (last modification: 28/05/2024)
		- Let's be honest, chatGPT 4 helped me on that one
	'''
	instances = {}
	def get_instance(*args, **kwargs):
		if cls not in instances:
			instances[cls] = cls(*args, **kwargs)
		return instances[cls]
	return get_instance


def is_numpy(val, shape = None, dtype = None):
	'''
	Determines if an input is a numpy array or not, eventually fitting some conditions about type and shape
	TODO: add dimension check
	Arguments:
		- val: the variable to test
		- shape: tuple the array.shape should return. Ignored if None.
		- dtype: the type array.dtype should belong to. Note that array.dtype can be a subtype of the target type (e.g. np.float32 is a subtype of np.floatings). Ignored if None.
	Returns:
		- A boolean telling if the input variable is a numpy array satisfying the required conditions
	Authors:
		- B.G. (last modification: 28/05/2024)
	'''
	
	# First checking if the value is a numpy array	
	if(isinstance(val, np.ndarray) == False):
		return False

	# Eventually checking its shape
	if(shape is not None and shape != val.shape):
		return False

	# Eventually checking its dtype
	if(dtype is not None and np.issubdtype(val.dtype, dtype) == False):
		return False

	# If the code reaches there then all good, the array is valid
	return True


def print_neighbourhood2D(arr, row, col, precision = 6):
	'''
	'''
	format_spec = f".{precision}f"
	gog = 0
	for i in [-1,0,1]:
		for j in [-1,0,1]:
			print('|', end = '')
			gog += 1
			if(gog == 3):
				gog = 0
				print(f"{ arr[ row + i , col + j ]:{format_spec}}", end = ' \n')
			else:
				print(f"{ arr[ row + i , col + j ]:{format_spec}}", end = ' | ')

	


'''
Set of internal functions to help computing gradients/surfaces from the combination of multiple fields/data
Some functions will be redundant but named differently for user friendliness 

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs


###############################################################
############# Surfaces ########################################
###############################################################


@ti.func
def Zw(Z: ti.template(), hw: ti.template(), i:ti.i32, j:ti.i32) -> ti.f32:
	'''
	Internal helping function returning the hydrayulic surface (elevation of the water surface)
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- i,j: the row col indices
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 30/04/2024)
	'''

	return Z[i,j] + ti.max(0.,hw[i,j])


@ti.func
def Zw_drape(Z: ti.template(), hw: ti.template(), i:ti.i32, j:ti.i32) -> ti.f32:
	'''
	Internal helping function returning the hydrayulic surface (elevation of the water surface)
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- i,j: the row col indices
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 30/04/2024)
	'''

	return Z[i,j] + hw[i,j]



###############################################################
############# Gradients #######################################
###############################################################

@ti.func
def Sw(Z: ti.template(), hw: ti.template(), i:ti.template(), j:ti.template(), ir:ti.template(), jr:ti.template())->ti.f32:
	'''
	Internal helping function returning the hydrayulic slope
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- i,j: the row col indices
		- ir,jr: the row col indices of the receivers node
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 20/05/2024)
	'''
	return (Zw(Z,hw, i,j) - Zw(Z,hw, ir,jr))/GRID.dx

@ti.func
def Sz(Z: ti.template(), i:ti.template(), j:ti.template(), ir:ti.template(), jr:ti.template())->ti.f32:
	'''
	Internal helping function returning the topographic slope
	Arguments:
		- Z: a 2D field of topographic elevation
		- i,j: the row col indices
		- ir,jr: the row col indices of the receivers node
	Returns:
		- the hydraulic surface
	Authors:
		- B.G. (last modification 20/05/2024)
	'''
	return (Z[i,j] - Z[ir,jr])/GRID.dx


@ti.func
def hydraulic_gradient_value(Z:ti.template(), hw:ti.template(),BCs:ti.template(),i:ti.i32, j:ti.i32 ) -> ti.f32:
	'''
	Calculates the local hydraulic gradient value.
	gradient = sqrt(max_slope_in_X^2 + max_slope_in_y^2)
	Convenient function, but only optimised for cases where calculating the hydraulic gradient value is the only neighbouring operation.

	Arguments:
		- Z: the field of topographic elevation
		- hw: the field of flow depth
		- BCs: the field of boundary conditions
		- i: the row index of the node in question
		- j: the column index of the node in question
	Returns:
		- the value of the steepest slope, 0 if no downslope neighbours
	Authors:
		- B.G. (last modification: 06/2024)
	'''


	# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
	SSx = 0.
	SSy = 0.
	gradSw = 0.

	# Traversing Neighbours
	for k in range(4):

		# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
		ir,jr = gridfuncs.neighbours(i,j,k, BCs)

		# if not a neighbours, by convention is < 0 and I pass
		if(ir == -1):
			continue

		if(gridfuncs.can_receive(ir,jr, BCs) == False):
			continue

		# Local hydraulic slope
		tS = Sw(Z,hw,i,j,ir,jr)

		# If < 0, neighbour is a donor and I am not interested
		if(tS <= 0):
			continue

		# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
		if(k == 0 or k == 3):
			if(tS > SSy):
				SSy = tS
		else:
			if(tS > SSx):
				SSx = tS

		# Done with processing this particular neighbour


	# Calculating local norm for the gradient
	# The condition manages the boundary conditions
	gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)

	return gradSw







'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib
import cmcrameri as cm

# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')

def str2cmap(tstr):
	'''
	Converts the string received from 
	'''
	if(tstr in plt.colormaps()):
		return tstr
	else:
		raise ValueError(f"{tstr} not a matplotlib colormap, if it is a cmcrameri cmap I am in the process of adding them");


class ColorMapWidget(QtWidgets.QComboBox):
	'''
		Colormap picker, a simple widget containing a list of colormap and sends a signals wiht the right colormap when it has changed
	'''

	# signal whenever if a new colormap is chosen is changed 
	colormapChanged = QtCore.Signal((str))

	def __init__(self):
		'''
		Simple widget in essence: only made of a list of options for colormaps and returns whichever is selected as a sring
		'''


		super(ColorMapWidget, self).__init__()
		# list of colormaps, to be completed
		self.addItems(['gist_earth', 'magma', 'viridis', 'cividis', 'Blues', 'Reds', 'RdBu'])
		# setting up the connection between inbuilt signal and my own
		self.currentIndexChanged.connect( lambda : self.colormapChanged.emit(self.currentText()) )




# end of file



'''
Riverdale's mirror for the numba engine (steenbock) convention for the nth neighbouring:

B.G - 07/2024 - Acigné

'''

import numba as nb
import numpy as np
from enum import Enum
import scabbard as scb


@nb.njit()
def mean_dist_to_outlet(Stack, Z, BCs, D8, nx, ny, dx):
	'''
	Internal computation of the mean flow distance to outlets

	Arguments:
		- Stack: topologically ordered nodes
		- Z: array of topography
		- BCs: array of boundary conditions codes
		- D8: bool, D8 if True else D4 flow routing
		- nx,ny,dx: dimensions and spatial steps
	Returns:
		- an array of flow distance from outlet

	Authors:
		- B.G. (last modifications 09/2024)
	'''

	# Init the distance to -1 as "not computed yet"
	dist = np.zeros_like(Z) - 1

	# Traversing the nodes in ascending order, downstream to upstream
	for node in Stack:

		# Checking if node is outlet - i.e. starting point
		if(scb.ste.can_out_flat(node,BCs)):
			dist[node] = 0

		# Checking if the node is valid to receive flow distance (e.g. can it give, is it no data)
		if(scb.ste.is_active_flat(node,BCs) == False) or (scb.ste.can_give_flat(node,BCs) == False):
			continue

		# prospective value
		val = 0.
		# Tracking number of values to average
		N = 0
		# For all neighbours (4 v. 8)
		for k in range(8 if D8 else 4):
			# get the neighbour
			rec = scb.ste.neighbours_D8_flat(node,k,BCs,nx,ny) if D8 else scb.ste.neighbours_D4_flat(node,k,BCs,nx,ny)
			# Check if valid (will be -1 if no data or cannot give)
			if(rec == -1):
				continue

			# Check edge case where internal unprocessed LM
			if(rec == node):
				continue
			# Check if actually is a receiver
			if(Z[node] <= Z[rec]):
				continue

			# Node is valid, increment number
			N +=1
			# incrementing the value
			val += dist[rec] + (scb.ste.dx_from_k_D8(dx,k) if D8 else dx)
		
		# end of loop and applying the mean
		if(N>0):
			dist[node] = val/N

	return dist

@nb.njit()
def min_dist_to_outlet(Stack, Z, BCs, D8, nx, ny, dx):
	'''
	Internal computation of the min flow distance to outlets

	Arguments:
		- Stack: topologically ordered nodes
		- Z: array of topography
		- BCs: array of boundary conditions codes
		- D8: bool, D8 if True else D4 flow routing
		- nx,ny,dx: dimensions and spatial steps
	Returns:
		- an array of flow distance from outlet

	Authors:
		- B.G. (last modifications 09/2024)
	'''

	# Init the distance to -1 as "not computed yet"
	dist = np.zeros_like(Z) - 1

	# Traversing the nodes in ascending order, downstream to upstream
	for node in Stack:

		# Checking if node is outlet - i.e. starting point
		if(scb.ste.can_out_flat(node,BCs)):
			dist[node] = 0

		# Checking if the node is valid to receive flow distance (e.g. can it give, is it no data)
		if(scb.ste.is_active_flat(node,BCs) == False) or (scb.ste.can_give_flat(node,BCs) == False):
			continue

		# prospective value
		val = 1e32
		# For all neighbours (4 v. 8)
		for k in range(8 if D8 else 4):
			# get the neighbour
			rec = scb.ste.neighbours_D8_flat(node,k,BCs,nx,ny) if D8 else scb.ste.neighbours_D4_flat(node,k,BCs,nx,ny)
			# Check if valid (will be -1 if no data or cannot give)
			if(rec == -1):
				continue

			# Check edge case where internal unprocessed LM
			if(rec == node):
				continue
			# Check if actually is a receiver
			if(Z[node] <= Z[rec]):
				continue

			# distance is the min of dist to that receiver and the other
			val = min(dist[rec] + (scb.ste.dx_from_k_D8(dx,k) if D8 else dx), val)

		dist[node] = val

	return dist


@nb.njit()
def max_dist_to_outlet(Stack, Z, BCs, D8, nx, ny, dx):
	'''
	Internal computation of the max flow distance to outlets

	Arguments:
		- Stack: topologically ordered nodes
		- Z: array of topography
		- BCs: array of boundary conditions codes
		- D8: bool, D8 if True else D4 flow routing
		- nx,ny,dx: dimensions and spatial steps
	Returns:
		- an array of flow distance from outlet

	Authors:
		- B.G. (last modifications 09/2024)
	'''

	# Init the distance to -1 as "not computed yet"
	dist = np.zeros_like(Z) - 1

	# Traversing the nodes in ascending order, downstream to upstream
	for node in Stack:

		# Checking if node is outlet - i.e. starting point
		if(scb.ste.can_out_flat(node,BCs)):
			dist[node] = 0

		# Checking if the node is valid to receive flow distance (e.g. can it give, is it no data)
		if(scb.ste.is_active_flat(node,BCs) == False) or (scb.ste.can_give_flat(node,BCs) == False):
			continue

		# prospective value
		val = 0.
		# For all neighbours (4 v. 8)
		for k in range(8 if D8 else 4):
			# get the neighbour
			rec = scb.ste.neighbours_D8_flat(node,k,BCs,nx,ny) if D8 else scb.ste.neighbours_D4_flat(node,k,BCs,nx,ny)
			# Check if valid (will be -1 if no data or cannot give)
			if(rec == -1):
				continue

			# Check edge case where internal unprocessed LM
			if(rec == node):
				continue
			# Check if actually is a receiver
			if(Z[node] <= Z[rec]):
				continue

			# distance is the max of dist to that receiver and the other
			val = max(dist[rec] + (scb.ste.dx_from_k_D8(dx,k) if D8 else dx), val)

		dist[node] = val

	return dist


import scabbard as scb
from nicegui import ui
import os
import numpy as np
from scabbard.riverdale.rd_params import param_from_dem
from scabbard.riverdale.rd_env import create_from_params, load_riverdale
import taichi as ti
import scabbard as scb
import matplotlib.pyplot as plt
import numpy as np
import time
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_hydrodynamics as hyd
import scabbard.riverdale.rd_hydrometrics as rdta
import scabbard.riverdale.rd_LM as lm
import scabbard.riverdale.rd_drainage_area as rda
from scabbard.riverdale.rd_hillshading import hillshading
import plotly.graph_objects as go
import cmcrameri.cm as cmc
import scabbard.nice_utils as nut
import scabbard.nice_helper as nhe
import sys



# initialising global stuff
# Dark mode on by default
dark = ui.dark_mode()
dark.enable()

# Initialising taichi also by default
ti.init(ti.gpu)

# Global variables
# callbacks rely on global variable retriavals 
# this dictionary just gathers all of them
stuff = {
	'ui':{},
	"fig" : None,
	"rd" : None,
	"colormap" : None,
	"im" : None,
	"color_range" : None,
	"value" : None,
	'model' : {"range": {"min": 0, "max": 100}},
}

# path of execution
CURRENT_PATH = os.getcwd()

def update_clim():
	'''
	Reinit the colorscale to the min/max of the plot
	'''
	global stuff
	nhe._update_clim(stuff)
	

def update_plot_value(stuff, val, cmin = None, cmax = None):
	'''
		Updates the plotted value: in the stuff, on the plots and the min-max of the colorscale
		Note: it does not call the plot.update function
	
	'''

	# Updating the actual values to plots, note the [::-1] to increase Y to the north
	stuff['value'] = val
	# Determining the min and max values
	tmin,tmax = np.nanmin(stuff['value']), np.nanmax(stuff['value'])
	if(cmin is not None):
		tmin = max(cmin, tmin)
	if(cmax is not None):
		tmax = min(cmax, tmax)
		
	# keeping in stuff
	stuff['model']['range']['min'] = tmin
	stuff['model']['range']['max'] = tmax

	# Updating the main figure
	stuff['main_figure'].update_traces(
		z=stuff['value'],
		zmin=stuff['model']['range']['min'],
		zmax=stuff['model']['range']['max'],
		selector=dict(name='datamap')
	)

	with stuff['ui']['r1c1']:
		ui.label('Colormap range')
		color_range = ui.range(min=stuff['model']['range']['min'], max=stuff['model']['range']['max'], 
			step = (stuff['model']['range']['max'] - stuff['model']['range']['min'])/250, 
			value = {'min': stuff['model']['range']['min'], 'max': stuff['model']['range']['max']}) \
		.props('label-always snap label-color="secondary" right-label-text-color="black"', ).bind_value(stuff['model'],'range').on('change', update_clim, throttle = 1)
		# # I'll need to make a factory for that 
	# ui.label('Colormap range')
	# color_range = ui.range(min=stuff['value'].min(), max=stuff['value'].max(), step = (stuff['value'].max() - stuff['value'].min())/250, value = {'min': 1000, 'max': 1400}) \
	# .props('label-always snap label-color="secondary" right-label-text-color="black"', ).bind_value(stuff['model'],'range').on('change', update_clim, throttle = 5)


def update_colorscale(stuff, cmap, title = "Colorbar"):

	with stuff['ui']['r1c1']:
		# Updating the main figure
		stuff['main_figure'].update_traces(
			colorscale=nut.cmap2plotly(cmap),colorbar=dict(title=title),
			selector=dict(name='datamap')
		)


def _update_tool_flowdepth():
	'''
	Sub-function updating the right column part of the tool to display the option for flow depth
	
	Arguments: None
	Returns: None
	Authors: None
	'''
	# Gathering the global stuff (lol)
	global stuff
	# Updating the value
	update_plot_value(stuff, stuff['rd'].hw.to_numpy()[::-1], cmin = 0)
	#Updating the colormap:
	update_colorscale(stuff, plt.get_cmap("Blues"), title = 'Flow Depth (m)')
	#redrawing hte plot
	stuff['ui']['plot'].update()

def _update_tool_topography():
	'''
	Sub-function updating the right column part of the tool to display the option for flow depth
	
	Arguments: None
	Returns: None
	Authors: None
	'''
	# Gathering the global stuff (lol)
	global stuff
	# Updating the value
	update_plot_value(stuff, stuff['rd'].Z.to_numpy()[::-1])
	#Updating the colormap:
	update_colorscale(stuff, plt.get_cmap("gist_earth"), title = 'Elevation (m)')
	#redrawing hte plot
	stuff['ui']['plot'].update()

def _update_tool_flow_velocity():
	'''
	Sub-function updating the right column part of the tool to display the option for flow depth
	
	Arguments: None
	Returns: None
	Authors: None
	'''
	# Gathering the global stuff (lol)
	global stuff
	# Updating the value
	update_plot_value(stuff, rdta.compute_flow_velocity(stuff['rd'])[::-1])
	#Updating the colormap:
	update_colorscale(stuff, plt.get_cmap("gist_earth"), title = 'u (m/s)')
	#redrawing hte plot
	stuff['ui']['plot'].update()

def _update_tool_shear_stress():
	'''
	Sub-function updating the right column part of the tool to display the option for flow depth
	
	Arguments: None
	Returns: None
	Authors: None
	'''
	# Gathering the global stuff (lol)
	global stuff
	# Updating the value
	update_plot_value(stuff, rdta.compute_shear_stress(stuff['rd'])[::-1])
	#Updating the colormap:
	update_colorscale(stuff, plt.get_cmap("magma"), title = 'Shear stress')
	#redrawing the plot
	stuff['ui']['plot'].update()

def _update_tool_a_eff():
	'''
	Sub-function updating the right column part of the tool to display the option for flow depth
	
	Arguments: None
	Returns: None
	Authors: None
	'''
	# Gathering the global stuff (lol)
	global stuff
	# Updating the value
	update_plot_value(stuff, rdta.compute_effective_drainage_area(stuff['rd'])[::-1])
	#Updating the colormap:
	update_colorscale(stuff, plt.get_cmap("cividis"), title = 'Effective drainage area')
	#redrawing the plot
	stuff['ui']['plot'].update()





update_tool_func = {
	'Topography': _update_tool_topography,
	'Flow depth': _update_tool_flowdepth,
	'Shear Stress': _update_tool_shear_stress,
	'Flow Velocity': _update_tool_flow_velocity,
	'Effective Area': _update_tool_a_eff,
}
def update_tool(event):
	'''
		This function is called when when the drop list is updated to a new too. It totally rewrite r1c1 with the new tool options
	'''
	global stuff
	# First, let's remove everything on that section
	stuff['ui']['r1c1'].clear()
	# Get the selected tool
	## Note: the try catch just depends on how the function is actually called
	try:
		tool = event['value']
	except:
		tool = event.value

	# Re-add the tool selector
	with stuff['ui']['r1c1']:
		stuff['ui']['tool_selector'] = ui.select([i for i in update_tool_func.keys()], value = tool)
		stuff['ui']['tool_selector'].on_value_change(update_tool)

	if update_tool_func[tool] is not None:
		update_tool_func[tool]()


async def pick_file() -> None:
	'''
	Load a file.
	Files supported so far:
		- RVD: the buit in riverdale's file format
		- TODO: load DEM from scratch and compute flow depth
	'''

	# just gathering the global dictionary
	global stuff
	# Getting the file name
	result = await scb.local_file_picker(CURRENT_PATH, multiple=False)
	ui.notify(f'Loading {result[0]}')
	# Actual loading operations
	stuff['rd'] = load_riverdale(result[0])
	stuff['value'] = stuff['rd'].Z.to_numpy()
	stuff['ui']['load_button'].delete()
	# Done with the loading
	ui.notify(f'Loaded !')

	# Setting up the Main GUI

	# Place holders:
	# Main row:
	stuff['ui']['r1'] =  ui.row()
	# Columns of main row
	with stuff['ui']['r1']:
		stuff['ui']['r1c0'] =  ui.column()
		stuff['ui']['r1c1'] = ui.column()

	stuff['main_figure'] = go.Figure()

	colorscale = nut.cmap2plotly(cmap = cmc.batlowK)

	# Add first image
	stuff['main_figure'].add_trace(go.Heatmap(
		z=stuff['rd'].Z.to_numpy()[::-1],
		colorscale=colorscale,
		zmin=1300,
		zmax=1320,
		zsmooth = 'best',
		# x0=0,
		# dx=500/A.shape[1],
		# y0=200,
		# dy=(800-200)/A.shape[0],
		colorbar=dict(title='elevation (m)'),
		name = 'datamap'
	))

	# Add second image
	stuff['main_figure'].add_trace(go.Heatmap(
		z=hillshading(stuff['rd'])[::-1],
		colorscale='gray',
		opacity=0.45,
		zsmooth = 'best',
		showscale=False,
		name = 'hillshade'
	))

	# Update layout
	stuff['main_figure'].update_layout(
		width=800,
		height=800,
		xaxis=dict(title='X'),
		yaxis=dict(title='Y'),
		margin=dict(l=20, r=20, t=20, b=20),

	)




	with stuff['ui']['r1c0']:
		stuff['ui']['plot'] = ui.plotly(stuff['main_figure']).classes('w-full h-40')

	# with stuff['ui']['r1c1']:
	update_tool({'value':'Topography'})



# First let's define the main layout of the app

stuff['ui']['r0'] = ui.row()

with stuff['ui']['r0']:
	ui.markdown('# GraphFlood - UI')
	ui.button('Quit', on_click=nut.quit_app)
	ui.button('Documentation', on_click=nut.notify_WIP)

stuff['ui']['load_button'] = ui.button('Load RVD file', on_click=pick_file, icon='folder')




# @ui.page('/')
# def index():
# with ui.row():
# 	ui.markdown('# Graphflood - Riverdale')
# 	ui.button('Quit', on_click=quit_app)
# load_button = ui.button('Choose file', on_click=pick_file, icon='folder')




if __name__ in {"__main__", "__mp_main__"}:
	ui.run()



import taichi as ti


@ti.kernel
def A_equals_B(A:ti.template(),B:ti.template()):
	for i,j in A:
		A[i,j] = B[i,j]


# Declare a function to calculate epsilon for a given value
@ti.func
def compute_epsilon_f32(value):
	eps = ti.cast(1.0, ti.f32)
	while ti.cast(value + eps, ti.f32) > value:
		eps /= 2.0
	return eps * 2.0


# Declare a function to calculate epsilon for a given value
@ti.func
def compute_epsilon_f64(value):
	eps = ti.cast(1.0, ti.f64)
	while ti.cast(value + eps, ti.f64) > value:
		eps /= 2.0
	return eps * 2.0


	


import json
import click
import pkg_resources
from importlib import resources 
import os

try:
	# Path to your JSON file
	# with resources.open_text('scabbard', 'config.json') as config_file:
		# config_fname = config_file.name
	# config_fname = pkg_resources.resource_filename('scabbard', 'data/config.json')
	config_fname = os.path.join(os.path.dirname(__file__), 'data', 'config.json')

except:
	print("Unable to read config")
	pass

# Path to your JSON file
def read_json(file_path):
	try:
		with open(file_path, 'r') as file:
			# Try to load the JSON content
			return json.load(file)
	except json.JSONDecodeError:
		# File is empty or not a valid JSON, return a default value
		# Return {} or [] depending on what your application expects
		return {}


def load_config():
	# Access the resource within the 'data' package/directory
	try:
		with open(config_fname,'r') as config_file:
			# print("config_file",config_file)
			try:
				config = json.load(config_file)
			except json.JSONDecodeError:
				config = {}
	except:
		print("Unable to read config")
		config = {}
	
	return config

@click.command()
def defaultConfig():

	print("Reintialising config to default")

	# data = {}
	# Reading the JSON file
	data = load_config()

	# Edit the data
	data['blender'] = '/home/bgailleton/code/blender/blender-4.0.1-linux-x64/blender'  # Modify this line according to your needs
	# data['blender'] = 'afslkdglasdfgsjldf'  # Modify this line according to your needs


	try:
		# Saving the modified data back to the JSON file
		with open(config_fname, 'w+') as file:
			json.dump(data, file, indent=4)
	except:
		print('unable to write config')

def query(paramstr = 'blender'):

	# Reading the JSON file
	data = load_config()

	return data[paramstr]





'''
Set of functions to generate specific arrays that I commonly use

B.G. (last modifications: 07/2024)
'''

import numpy as np

def normal_BCs_from_shape(nx,ny, out_code = 3):
	'''
	Returns an array on uint8 boundary codes with "normal" edges

	Arguments:
		- nx (int): number of columns
		- ny (int): number of rows
		- out_code (uint8): code for the edges, 3 is "can_out", a permissive code to let flow out if no downstream neighbour

	returns:
		- a 2D numpy array of boundary conditions
	
	Authors:
		- B.G. (last modification 07/2024)
	'''

	BCs = np.ones((ny,nx), dtype = np.uint8)
	BCs[:,[-1,0]] = out_code
	BCs[[-1,0], :] = out_code

	return BCs



"""Top-level package for scabbard."""

__author__ = """Boris Gailleton"""
__email__ = 'boris.gailleton@univ-rennes.fr'

import sys
import platform

# Dictionary to map the OS and Python version to the wheel file
wheel_urls = {
	'Linux': {
		'310': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp310-cp310-linux_x86_64.whl',
		'311': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp311-cp311-linux_x86_64.whl',
		'312': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp312-cp312-linux_x86_64.whl'
	},
	'Darwin': {  # macOS
		'310': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp310-cp310-macosx_14_0_arm64.whl',
		'311': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp311-cp311-macosx_14_0_arm64.whl',
		'312': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp312-cp312-macosx_14_0_arm64.whl'
	},
	'Windows': {
		'310': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp310-cp310-win_amd64.whl',
		'311': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp311-cp311-win_amd64.whl',
		'312': 'https://github.com/bgailleton/scabbard/raw/refs/heads/main/wheels/pytopotoolbox/topotoolbox-3.0.1-cp312-cp312-win_amd64.whl'
	}
}

def suggest_installation():
	python_version = f"{sys.version_info.major}{sys.version_info.minor}"
	os_name = platform.system()

	print('As of today, the "official" installation procedure is Work In Progress, so I precompiled the binaries but you need to install them (couple of command lines max)')

	# Match OS names to dictionary keys
	if os_name == 'Darwin':
		os_key = 'Darwin'
	elif os_name == 'Windows':
		os_key = 'Windows'
	elif os_name == 'Linux':
		os_key = 'Linux'
	else:
		print(f"Unsupported OS: {os_name}")
		return

	# Get the appropriate wheel link
	wheel_link = wheel_urls.get(os_key, {}).get(python_version)

	if wheel_link:
		print(f"To install the appropriate wheel, run:")
		print(f"pip install {wheel_link}")
	else:
		print(f"Sorry, no wheel available for Python {python_version} on {os_name}.")





TTB_AVAILABLE = False

try:
	# Common import centralised
	import topotoolbox as ttb
	TTB_AVAILABLE = True
except:
	import os
	# Call the function when the module is imported
	suggest_installation()
	raise RuntimeError('''
pytopotoolbox is required for using scabbard. 
See above for installation instruction.
''')

# Legacy imports
from .config import *
from .enumeration import *
from .utils import *
from .shape_functions import *
from .lio import *
from .fastflood import *
from .geography import *
from .grid import *
from .Dax import *
from .Dfig import *
from .Dplot import *
from .graphflood import *
from .phineas import *
from .graphflood_helper import *
from .environment import *
from .blendplot import *
# from .local_file_picker import *


# New module-type import system
from . import _utils as ut
from . import raster
from . import riverdale
from . import steenbok
from . import riverdale as rvd
from . import steenbok as ste
from . import filters
from . import flow
from . import visu
from . import io
from . import graphflood_ui as graphflood






'''
Contains sets of utils function to manipulate arrays
'''

from scipy.ndimage import zoom, label
import numpy as np
import numba as nb


def resample_to_shape(array, new_shape, order = 3):
	"""
	Resample a 2D array to a new shape.
	
	Parameters:
		- array (2D numpy array): The original array to be resampled.
		- new_shape (tuple): The desired shape (rows, columns) for the resampled array.
	
	Returns:
		- 2D numpy array: The resampled array with the specified shape.
	
	Authors: 
		- B.G.
		- a LLM
	"""
	# Calculate the zoom factors for each dimension
	zoom_factors = (new_shape[0] / array.shape[0], new_shape[1] / array.shape[1])
	
	# Use scipy.ndimage.zoom to resample the array
	resampled_array = zoom(array, zoom_factors, order=order)
	
	return resampled_array


def remove_unconnected_components(mask, th_components = 10, D8 = True):
	'''
	Edit a mask in place: removes (switch to 0) components that are isolated/in patches of less than a threshold

	Arguments:
		- mask: 2D numpy array of 0s and 1s
		- the minimum number of connected component
		- D8: if True, takes diagonals into account, else only cardinals

	Returns:
		Nothing, edits mask in place 
	'''

	# Nested function to apply the removal (much faster than numpy)
	@nb.njit
	def _remover_for_remove_unconnected_components(mask, labels, N, th_components):

		Ns = np.zeros(N, dtype = np.uint32)

		for i in range(labels.shape[0]):
			for j in range(labels.shape[1]):
				Ns[labels[i,j]] += 1

		for i in range(labels.shape[0]):
			for j in range(labels.shape[1]):
				if(Ns[labels[i,j]] <= th_components):
					mask[i,j] = 0


	# defines the neighbouring
	structure = [[0,1,0],[1,1,1],[0,1,0]] if D8 == False else [[1,1,1],[1,1,1],[1,1,1]]

	# connected components
	labeled_array, num_features = label(mask, structure = structure)
	
	# correct the mask
	_remover_for_remove_unconnected_components(mask, labeled_array, num_features, th_components)




import numpy as np
import matplotlib.pyplot as plt
import scabbard.visu.base as base
import scabbard as scb
import click
import os

@click.command()
@click.option('-s', '--sea_level', 'sea_lvl',  type = float, default = 0.)
@click.argument('fname', type = str)
def cli_nice_terrain(fname, sea_lvl):
	
	if(os.path.exists(fname) == False):
		raise ValueError(f'FileDoesNotExists: {fname}')
	
	dem = scb.io.load_raster(fname)
	
	fig,ax = base.hs_drape(dem, dem.Z, cmap = 'terrain', label = 'Elevation', alpha = 0.4, 
	cut_off_min = None, cut_off_max = None, sea_level = sea_lvl, vmin = None,
	vmax = None, res = None, fig = None, ax = None)
	
	plt.show()

def nice_terrain(dem, cmap = 'terrain', alpha = 0.55, 
	sea_level = None, vmin = None,
	vmax = None, res = None, mask = None, **kwargs):
	'''
	Quick visualisation of a DEM as a hillshade + an imshow-like data on the top of it with the same extent

	Arguments:
		- dem: the topographic grid
		- arr2D: the array to be draped, to the extent of the dem
		- cmap: the colormap to use
		- label: label of the colorbar
		- alpha: the transparency of the arr2D
		- cut_off_min/max: any data on arr2D below or above this value will be ignored
		- sea_level: any data on topo and arr2D where topo < that value will be ignored
		- vmin/vmax: the extents of the colors for the cmap of arr2D
		- mask: a binary mask to set additional nodata sections manually
		- **kwargs: anything that goes into fig, ax = plt.subplots(...)
	Returns:
		- fig,ax  with everything plotted on it

	Author:
		- B.G. (last modification: 08/2024)
	'''

	fig,ax = base.hs_drape(dem, (dem.Z if isinstance(dem,scb.raster.RegularRasterGrid) else dem.Z2D), cmap = cmap, label = 'Elevation', alpha = alpha, 
		sea_level = sea_level, vmin = vmin, vmax = vmax, mask = mask, **kwargs)

	return fig,ax

